/***************************************************************************

    file                 : driver.cpp
    created              : Thu Dec 20 01:21:49 CET 2002
    copyright            : (C) 2002-2004 Bernhard Wymann
    email                : berniw@bluewin.ch
    version              : $Id: driver.cpp,v 1.16 2006/04/27 22:32:29 berniw Exp $

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//#define DRV_DEBUG 

#include <string>
#include <tgfclient.h>

#include "driver.h"
#include "vardef.h"



// Static variables.
static int pitstatus[128] = {0};
Cardata *Driver::cardata = NULL;


Driver::Driver(int index)
{
	INDEX = index;
	//myCarName = CAR_NAME;
	moduleName = BOT_NAME;
	
/* Constants */
G = 9.81f;					// [m/(s*s)] Welcome on Earth.
MAX_UNSTUCK_ANGLE = (float)(15.0f/180.0f*PI);	// [radians] If the angle of the car on the track is smaller, we assume we are not stuck.
UNSTUCK_TIME_LIMIT = 3.0f;			// [s] We try to get unstuck after this time.
UNSTUCK2_TIME_LIMIT = 9.0f;			// [s] We try to get unstuck after this time.
UNSTUCK3_TIME_LIMIT = 18.0f;			// [s] We try to get unstuck after this time.
MAX_UNSTUCK_SPEED = 5.0f;			// [m/s] Below this speed we consider being stuck.
MIN_UNSTUCK_DIST = 1.0f;			// [m] If we are closer to the middle we assume to be not stuck. <2.0>

FULL_ACCEL_MARGIN = 1.0f;			// [m/s] Margin reduce oscillation of brake/acceleration.
SHIFT = 0.98f;					// [-] (% of rpmredline) When do we like to shift gears. <0.96>
SHIFT_UP = 0.99f;				// [-] (% of rpmredline) 
SHIFT_DOWN = 120;
SHIFT_MARGIN = 4.0f;				// [m/s] Avoid oscillating gear changes.

ABS_SLIP = 2.5f;				// [m/s] range [0..10]
ABS_RANGE = 5.0f;				// [m/s] range [0..10]
ABS_MINSPEED = 3.0f;				// [m/s] Below this speed the ABS is disabled (numeric, division by small numbers).
TCL_SLIP = 2.0f;				// [m/s] range [0..10]
TCL_RANGE = 10.0f;				// [m/s] range [0..10]

LOOKAHEAD_CONST = 12.0f;			// [m]
LOOKAHEAD_FACTOR = 0.33f;			// [-]
WIDTHDIV = 2.0f;				// [-] Defines the percentage of the track to use (2/WIDTHDIV). <3.0>
SIDECOLL_MARGIN = 3.0f;				// [m] Distance between car centers to avoid side collisions. 
FRONTCOLL_MARGIN = 4.0f;			// [m] Distance between car to avoid collisions.
BORDER_OVERTAKE_MARGIN = 0.5f;			// [m]
OVERTAKE_OFFSET_SPEED = 5.0f;			// [m/s] Offset change speed.
PIT_LOOKAHEAD = 6.0f;				// [m] Lookahead to stop in the pit.
PIT_BRAKE_AHEAD = 200.0f;			// [m] Workaround for "broken" pitentries.
PIT_MU = 0.4f;					// [-] Friction of pit concrete.
MAX_SPEED = 84.0f;				// [m/s] Speed to compute the percentage of brake to apply.
MAXFUEL_PERMETER = 0.0012f;			// [liter/m] fuel consumtion.
CLUTCH_SPEED = 5.0f;				// [m/s]
CENTERDIV = 0.1f;				// [-] (factor) [0.01..0.6].
DISTCUTOFF = 200.0f;				// [m] How far to look, terminate while loops.
MAX_INC_FACTOR = 6.0f;				// [m] Increment faster if speed is slow [1.0..10.0].
CATCH_FACTOR = 10.0f;				// [-] select MIN(catchdist, dist*CATCH_FACTOR) to overtake.
CLUTCH_FULL_MAX_TIME = 2.0f;			// [s] Time to apply full clutch.
USE_LEARNED_OFFSET_RANGE = 0.2f;		// [m] if offset < this use the learned stuff

TEAM_REAR_DIST = 50.0f;				//
MAX_DAMAGE_DIFF = 1000;

	mCar = NULL;
	mTrack = NULL;
  	mSitu = NULL;
	stuckTimer = false;
	stucksteer = -20.0;
	isStuck = false;
	count_Stuck = 0;
	maxCarSpeed = 0.0;
  	minCarSpeed = 300.0;
	car_damage = 0;
  	car_damageLast = 0;
  	maxCarDamage = 0;
	MU_FACTOR = 0.69f;
	FUEL_FACTOR = 1.5;
	maxDamage = 3500;
	brakedelay = 10.0;
	PitOffset = 10.0;
	LetPass = 0.5;
	bumpCaution = 0;
	setAccel = 0;
	CornerSpeed = 15.0;
	CARMASS = 1150.0;
	ftank_Mass = 85.0;
	car_Mass = 1250.0;
}

Driver::~Driver()
{
	delete opponents;
	delete pit;
	delete [] radius;
	delete strategy;
	if (cardata != NULL) {
		delete cardata;
		cardata = NULL;
	}
}


/*=========================================================*/

// Called for every track change or new race.
void Driver::initTrack(tTrack* t, void *carHandle, void **carParmHandle, tSituation *s)
{
	track = t;
	mTrack = t;
	mSitu = s;

//	const int BUFSIZE = 256;
//	char buffer[BUFSIZE];
	/*------------------------------------------------------*/
	/*	 Load a custom setup if one is available.	*/
	/*------------------------------------------------------*/
	//	get the name of the track.
	//mytrackName = track->internalname;
	// 	Get a pointer to the first char of the track filename.
//  	char* trackName = strrchr(track->filename, '/') + 1;
	char trackName[256];
	strncpy( trackName, strrchr(track->filename, '/') + 1, sizeof(trackName) );
	*strrchr(trackName, '.') = '\0'; // remove suffix (.xml)

	void* newParmHandle = NULL;
	*carParmHandle = NULL;
	char buf[1024];

	// 	set up params for all cars and all tracks.
    	snprintf( buf, sizeof(buf), "drivers/%s/setup/setup.xml", moduleName );
	*carParmHandle = GfParmReadFile(buf, GFPARM_RMODE_STD);
	if (*carParmHandle != NULL) {
		newParmHandle = *carParmHandle;
	} else {
    		saveSegData = false;
    		saveRaceData = false;
    		showSegIndex = 0;
    		drv_verbose = 0;
    		mode_verbose = 0;
		opp_verbose = 0;
    		strat_verbose = 0;
    		testPitstop = 0;
    		testQualifTime = 0;
		drv_setup = false;
	}

	//	get the name of the car:
  /*
	// in the XML file
   	 const char *teamcar = GfParmGetStr(*CarParmHandle, SECT_PRIV, "carname", NULL);
	if (teamcar != NULL) {
		myCarName = teamcar;
	} else {
		myCarName = CAR_NAME;
	}
  */
	// from the Torcs car.h ( #define PRM_CAR   "car" )
	char*	path = SECT_GROBJECTS "/" LST_RANGES "/" "1";
	char*	key  = PRM_CAR;
	//char	carName[256];
	strncpy( carName, GfParmGetStr(carHandle, path, key, ""), sizeof(carName) );
	char*	p = strrchr(carName, '.');
	if( p ) *p = '\0';
	myCarName = carName;


	*carParmHandle = NULL;
	//	set up race type array.
	char*	raceType[] = { "practice", "qualif", "race" };

	// 	set up params for car type on track of specific race type.
    	snprintf( buf, sizeof(buf), "drivers/%s/cars/%s/%s-%s.xml",
				moduleName, carName, trackName, raceType[s->_raceType] );
	*carParmHandle = GfParmReadFile(buf, GFPARM_RMODE_STD);

	if (*carParmHandle == NULL) {
	// 	set default params for this car if no race type params defined
    	snprintf( buf, sizeof(buf), "drivers/%s/cars/%s/default.xml", moduleName, carName );
	*carParmHandle = GfParmReadFile(buf, GFPARM_RMODE_STD);;
	}

	// 	merge the car param handle.
	if (*carParmHandle != NULL && newParmHandle !=NULL) {
     	*carParmHandle = GfParmMergeHandles(*carParmHandle, newParmHandle, (GFPARM_MMODE_SRC|
					GFPARM_MMODE_DST|GFPARM_MMODE_RELSRC|GFPARM_MMODE_RELDST));
	}

	GfOut( "## initTrack(): moduleName %s carName %s trackName %s raceType %s\n", 
			moduleName, carName, trackName, raceType[s->_raceType] );

  /*-----------------------------------------------------*/
  /*  			Define a Strategy		 */
  /*-----------------------------------------------------*/
   // Simple management of fuel and repairs
	strategy = new SimpleStrategy();
	strategy->setFuelAtRaceStart(t, carHandle, carParmHandle, s, INDEX);

		/* Load data from the xml file and set parameters. */
	MU_FACTOR = GfParmGetNum(*carParmHandle, SECT_PRIV, PRV_ATT_MUFACTOR, (char*)NULL, 0.69f);
	CHANGE_GEAR = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, PRV_ATT_CHANGEGEAR, (char*)NULL, 1);
	MAX_GEAR = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, PRV_ATT_MAXGEAR, (char*)NULL, 6);
	fuel_PerMeter = GfParmGetNum(*carParmHandle, SECT_PRIV, PRV_ATT_FUELPERMETER, (char*)NULL, MAXFUEL_PERMETER);
	fuel_PerLap = GfParmGetNum(*carParmHandle, SECT_PRIV, PRV_ATT_FUELPERLAP, (char*)NULL, track->length * fuel_PerMeter);
	saveSegData = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "savesegdata", (char*)NULL, 0.0);
    	saveRaceData = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "saveracedata", (char*)NULL, 0.0);
    	drv_verbose = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "driververbose", (char*)NULL, 0.0);
    	mode_verbose = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "modeverbose", (char*)NULL, 0.0);
	opp_verbose = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "oppverbose", (char*)NULL, 0.0);
    	strat_verbose = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "strategyverbose", (char*)NULL, 0.0);
    	testPitstop = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "chkpitstop", (char*)NULL, 0.0);
    	testQualifTime = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "chkqualiftime", (char*)NULL, 0.0);
    	showSegIndex = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "segindex", (char*)NULL, 0.0);
	drv_setup = (int)GfParmGetNum(*carParmHandle, SECT_PRIV, "showsetting", (char*)NULL, 0.0);
	LetPass = GfParmGetNum( *carParmHandle, SECT_PRIV, "LetPass", (char*)NULL, 0.5 );
	tcl_slip = GfParmGetNum( *carParmHandle, SECT_PRIV, "tcl_slip", (char*)NULL, TCL_SLIP );
	tcl_range = GfParmGetNum( *carParmHandle, SECT_PRIV, "tcl_range", (char*)NULL, TCL_RANGE );
	abs_slip = GfParmGetNum( *carParmHandle, SECT_PRIV, "abs_slip", (char*)NULL, ABS_SLIP );
	abs_range = GfParmGetNum( *carParmHandle, SECT_PRIV, "abs_range", (char*)NULL, ABS_RANGE );
	maxDamage = GfParmGetNum( *carParmHandle, SECT_PRIV, PRV_ATT_MAXDAMAGE, (char *)NULL, 1.0f );
	pitStrat = GfParmGetNum( *carParmHandle, SECT_PRIV, PRV_ATT_PITSTRAT, (char *)NULL, 1.0f );
	PitOffset = GfParmGetNum( *carParmHandle, SECT_PRIV, PRV_ATT_PITOFFSET, (char *)NULL, 10.0 );
	pitEntrySpeed = GfParmGetNum(*carParmHandle, SECT_PRIV, PRV_ATT_PITENTRY, (char*)NULL, 25.0);
  	pitExitSpeed = GfParmGetNum(*carParmHandle, SECT_PRIV, PRV_ATT_PITEXIT, (char*)NULL, 25.0);
	pitExitMargin = GfParmGetNum(*carParmHandle, SECT_PRIV, PRV_ATT_PITEXITMARGIN, (char*)NULL, 80.0);
	FUEL_FACTOR = GfParmGetNum(*carParmHandle, SECT_PRIV, PRV_ATT_FUELTANKFACTOR, (char*)NULL, 1.0);
        //BrakeFactor = GfParmGetNum( *carParmHandle, SECT_PRIV, PRV_ATT_BRAKEFACTOR, (char *)NULL, 0.33f );
        //SteerDampingFactor = GfParmGetNum( *carParmHandle, SECT_PRIV, "steerdampingfactor", (char *)NULL, 0.5f );
	//brakedistfactor = GfParmGetNum( *carParmHandle, SECT_PRIV, "brakedistfactor", (char *)NULL, 1.6f );
	brakedelay = GfParmGetNum( *carParmHandle, SECT_PRIV, "BrakeDelay", (char *)NULL, 10.0 );
	CornerSpeed = GfParmGetNum( *carParmHandle, SECT_PRIV, "CornerSpeed", (char *)NULL, 15.0 );
	double IntMargin = GfParmGetNum( *carParmHandle, SECT_PRIV, "IntMargin", (char *)NULL, 1.0 );
	double ExtMargin = GfParmGetNum( *carParmHandle, SECT_PRIV, "ExtMargin", (char *)NULL, 2.0 );

	/* K1999 raceline */
	raceline.setCornerSpeed( CornerSpeed );
	raceline.setBrakeDist( brakedelay );
	raceline.setIntMargin( IntMargin );
	raceline.setExtMargin( ExtMargin );
	raceline.setRwData( track, carParmHandle, s );
}

/*=========================================================*/
// Start a new race.
void Driver::newRace(tCarElt* car, tSituation *s)
{
	float deltaTime = (float) RCM_MAX_DT_ROBOTS;
	MAX_UNSTUCK_COUNT  = int(UNSTUCK_TIME_LIMIT/deltaTime);
	MAX_UNSTUCK_COUNT2 = int(UNSTUCK2_TIME_LIMIT/deltaTime);
	MAX_UNSTUCK_COUNT3 = int(UNSTUCK3_TIME_LIMIT/deltaTime);
	OVERTAKE_OFFSET_INC = OVERTAKE_OFFSET_SPEED*deltaTime;
	lastlookahead = laststeer = lastNSasteer = 0.0f;
	stuck = 0;
	alone = 1;
	clutchtime = 0.0f;
	myoffset = 0.0f;
	simtime = correcttimer = 0.0;
	correctlimit = 1000.0;
	this->car = car;
	mCar = car;
	mSitu = s;

	// set car specifications
	initCa();
	initCw();
  	initCarSpec();
	initCR();
	initTireMu();
	initTCLfilter();
	if (drv_setup) {
		showSetup();
	}

	// Create just one instance of cardata shared by all drivers.
	if (cardata == NULL) {
		cardata = new Cardata(mSitu);
	}
	mycardata = cardata->findCar(mCar);
	currentsimtime = mSitu->currentTime;

	// initialize the list of opponents.
	opponents = new Opponents(mSitu, this, cardata);
	opponent = opponents->getOpponentPtr();

	// create the pit object.
	pit = new Pit(mSitu, this, PitOffset);
	setMode( mode_correcting );
	lastmode = mode_correcting;
	if (mode_verbose)
		fprintf(stderr,"%s enter in correcting mode\n", mCar->_name);
	carindex = 0;

	for (int i = 0; i < mSitu->_ncars; i++)
	{
		if (mSitu->cars[i] == mCar)
		{
			carindex = i;
			break;
 		}
 	}

	/* K1999 raceline */
	raceline.setCar( mCar );
	raceline.NewRace( mCar, mSitu );
	raceline.InitTrack(track, mSitu);

	line.setCar( mCar );
	line.InitTrack(track, mSitu);

	// Initialize radius of segments.
	radius = new float[track->nseg];
	//computeRadius( ref_line, radius);
	computeRadius( 2, radius);
}

/*=========================================================*/
// Drive during race.
void Driver::drive()
{	
  	clearCtrl();
	updateTime();
	update();
 // 	drvUpdate();
	stateStuck();
	if (isStuck) {
	        stucksteer =  -mycardata->getCarAngle() / mCar->_steerLock;
		if ( drv_verbose )
			fprintf(stderr,"# %s is Stuck!\n", mCar->_name);
	}
	calcSpeed();
	setControls();
	laststeer = mCar->_steerCmd;
	lastmode = Mode;

}

/*=========================================================*/
// Set pitstop commands.
int Driver::pitCommand()
{
	mCar->_pitRepair = strategy->pitRepair(mCar, mSitu);
	mCar->_pitFuel = strategy->pitRefuel(mCar, mSitu);
	// This should be the only place where the pit stop is set to false!
	pit->setPitstop(false);
	return ROB_PIT_IM; // return immediately.
}

/*=========================================================*/
// End of the current race.
void Driver::endRace(tSituation *s)
{
	// Nothing for now.
}

/***************************************************************************
 *
 * 			utility functions
 *
***************************************************************************/

// Init Car specifications
void Driver::initCarSpec()
{
	CARMASS = GfParmGetNum(mCar->_carHandle, (char *)SECT_CAR, (char *)PRM_MASS, NULL, 1150.0f);
	double wheelFRONT = GfParmGetNum(mCar->_carHandle, SECT_FRNTAXLE, PRM_XPOS, (char*)NULL, 0.0);
	double wheelREAR = GfParmGetNum(mCar->_carHandle, SECT_REARAXLE, PRM_XPOS, (char*)NULL, 0.0);
  	WHEEL_BASE =  wheelFRONT - wheelREAR;
  	BRAKE_PRESSURE = GfParmGetNum(mCar->_carHandle, SECT_BRKSYST, PRM_BRKPRESS, NULL, 20000000.0) / 1000.0;
}

void Driver::clearCtrl()
{
	memset(&mCar->ctrl, 0, sizeof(tCarCtrl));
}

// Update my private data every timestep.
void Driver::update()
{
	opponents->update(mSitu, this);
	strategy->update(mCar, mSitu);
	posFromStart = mCar->_distFromStartLine;

	if (mCar->_state <= RM_CAR_STATE_PIT)
	{
		//if (!pit->getPitstop() && !pit->isBetween(posFromStart))
		if (!pit->getPitstop() && (posFromStart < pit->getNPitEntry() || posFromStart > pit->getNPitEnd())) 
		{
			pit->setPitstop(strategy->needPitstop(mCar, mSitu));
		}
		setPitStatus();
	} else pitstatus[carindex] = 0;

	pit->update();
	alone = isAlone();
	simtime = mSitu->currentTime;
	carSpeed = mCar->_speed_x;
	updateData();
	
	/*      		car3-trb1 :
	   body length = 4.55m, body width = 1.95m, body height = 1.12m
	*/
	//double car_width = 1.94 / 2.0;
	//double car_length = 4.54 / 2.0;
	float car_width = mCar->_dimension_y / 2.0;
	float car_length = mCar->_dimension_x / 2.0;
	
	// car pos/state on the track; check distance to border/barrier.
	double carToMiddle = mCar->_trkPos.toMiddle;
  	bool car_onLeftSide = carToMiddle > 0.0 ? true : false;
  	carBorderdist = mCar->_trkPos.seg->width / 2.0 - fabs(carToMiddle) - car_width;
  	carWalldist = mCar->_trkPos.seg->width / 2.0 - fabs(carToMiddle) - car_length;
  	if (mCar->_trkPos.seg->side[car_onLeftSide] != NULL) {
    		carWalldist += mCar->_trkPos.seg->side[car_onLeftSide]->width - 1.0;
    		if (mCar->_trkPos.seg->side[car_onLeftSide]->side[car_onLeftSide] != NULL) {
      			carWalldist += mCar->_trkPos.seg->side[car_onLeftSide]->side[car_onLeftSide]->width;
    		}
  	}

	float angleToTrack = RtTrackSideTgAngleL(&(mCar->_trkPos)) - mCar->_yaw;
	NORM_PI_PI(angleToTrack);
	bool angle_ToLeft = angleToTrack < 0.0 ? true : false;
  	car_goToWall = (angle_ToLeft == car_onLeftSide) ? true : false;
	angle = -angleToTrack;

// update car damage to change lead and PitStop.
	car_damage = mCar->_dammage;
  	car_damageDiff = mCar->_dammage - car_damageLast;
  	car_damageLeft = maxDamage - mCar->_dammage;
  	car_damageLast = mCar->_dammage;

	updateStuck();
}

void Driver::updateTime()
{
  	//currentsimtime = mSitu->currentTime;
	if (currentsimtime != mSitu->currentTime) {
		currentsimtime = mSitu->currentTime;
		cardata->update();
	}
	// update Timer (to update Stuck)
	double Time_Diff;
  	Time_Diff = currentsimtime - lastTimer;
  	if (Time_Diff >= 0.1) {
    		lastTimer += 0.1;
    		stuckTimer = true;
  	} else {
   		 stuckTimer = false;
  	}
}


// Update global car data.
void Driver::updateData()
{
	// Update the local data rest.
	speedangle = (float) -(mycardata->getTrackangle() - atan2(mCar->_speed_Y, mCar->_speed_X));
	NORM_PI_PI(speedangle);
	ftank_Mass = mCar->_fuel * FUEL_FACTOR;
	car_Mass = CARMASS + ftank_Mass;
	carSpeed = mCar->_speed_x;
	currentspeedsqr = carSpeed * carSpeed;
  	if (mCar->race.laps > 0) {
  		if (carSpeed > maxCarSpeed) maxCarSpeed = carSpeed;
  		if (carSpeed < minCarSpeed) minCarSpeed = carSpeed;
  		if (car_damage > maxCarDamage) maxCarDamage = car_damage;
  	}
	//carMu = mCar->_trkPos.seg->surface->kFriction;
	carMu = (mCar->_trkPos.seg->surface->kFriction * TIREMU) * MU_FACTOR;
  	carFriction = carMu * (car_Mass * G + CA * currentspeedsqr);

	if ( drv_verbose ) 
		GfOut("# %s: MinSpeed %.2f MaxSpeed %.2f CarMass %.2f CarFriction %f\n", 
			mCar->_name, minCarSpeed, maxCarSpeed, car_Mass, carFriction);
}

void Driver::setPitStatus()
{
	if (pit->getPitstop() && mCar->_pit) {
		pitstatus[carindex] = 1;
		for (int i=0; i<opponents->getNOpponents(); i++) {
			int idx = opponent[i].getIndex();
			if (opponent[i].getTeam() != TEAM_MATE) continue;
			if (opponent[i].getCarPtr() == mCar) continue;
			if (opponent[i].getCarPtr()->_state > RM_CAR_STATE_PIT) continue;
			if (pitstatus[idx] == 1 ||
				    ((pitstatus[idx] || (opponent[i].getCarPtr()->_fuel < mCar->_fuel-1.0 && mCar->_dammage < 5000))
				    && fabs(mCar->_trkPos.toMiddle) <= mCar->_trkPos.seg->width/2))
			{
				pit->setPitstop( 0 );
				pitstatus[carindex] = 0;
			}
			break;
		}
	} else if (!pit->getInPit()) pitstatus[carindex] = 0;

	return;
}

void Driver::stateStuck()
{
	if (fabs(mycardata->getCarAngle()) > MAX_UNSTUCK_ANGLE && mCar->_speed_x < MAX_UNSTUCK_SPEED &&
	    fabs(mCar->_trkPos.toMiddle) > MIN_UNSTUCK_DIST) {
	    if ( stuck > MAX_UNSTUCK_COUNT3 ) {
		/* if we're stuck for really long and not moving, try going again */
		stuck = 0;
		isStuck = false;
	    } 
	    else if (stuck > MAX_UNSTUCK_COUNT && mCar->_trkPos.toMiddle*mycardata->getCarAngle() < 0.0) {
		stuck++;
		isStuck = true;
	    } else {
		stuck++;
		isStuck = false;
	    }
	} else if ( fabs(mCar->_speed_x) < MAX_UNSTUCK_SPEED ) {
	    /* not moving == stuck */
	    if ( stuck > MAX_UNSTUCK_COUNT3 ) {
		/* if we're stuck for really long and not moving, try going again */
		stuck = 0;
		isStuck = false;
	    } else if ( stuck > MAX_UNSTUCK_COUNT2 ) {
		stuck++;
		isStuck = true;
	    } else {
		stuck++;
		isStuck = false;
	    }
	} else {
	    stuck = 0;
	    isStuck = false;
	}
}

void Driver::updateStuck()
{
 /* check for why the car is waiting! */
	isWait = false;
  	onColl = false;
	double speed = mCar->_speed_x;
	double carBrakeDist = 0.0;
	if ( speed > 0.0 ) {
		carBrakeDist = (car_Mass * currentspeedsqr / 2.0) / (carMu * G * car_Mass * (CA * carMu + CW));
	}

  if (mCar->race.laps > 0) {
	if (car_goToWall && fabs(angle) > 0.7) {
		if (carWalldist < carBrakeDist && !isStuck) {
     			 onColl = true;
    		}
	}
//
	if (carSpeed <= 0.0 && !isStuck) {
     		 onColl = true;
		if (  mCar->_gear == 0 ) {
			isWait = true;
		}
	}
//
	if ( !car_goToWall && mCar->_gear <= 0 ) {
		onColl = true;
		isWait = true;
	}
  }
 /* TODO
  Maybe now check opponents distance and speed...
   use onColl to set new Mode (mode_stuck)
   and update braking and speed.
 ------------------------------------------- */

  	if (stuckTimer) {
    		if (isWait) {
      			count_Stuck = 0;
    		}
    		if (isStuck) {
      			if (fabs(carSpeed) < 7.0) {
        			if (count_Stuck++ > 20) {
          				count_Stuck = 0;
          				isStuck = false;
        			}
      			} else {
        			count_Stuck = 0;
        			isStuck = false;
      			}
    		} else if (fabs(carSpeed) < 0.5) {
      			if (count_Stuck++ > 10) {
        			count_Stuck = 0;
        			isStuck = true;
      			}
    		} else {
      			count_Stuck = 0;
    		}
  	}
}

void Driver::setAccelBrake(double speed, double barkefactor)
{
	double x = (10 + mCar->_speed_x) * (speed - mCar->_speed_x) / 200;
	if (x > 0) accelcmd = (float) x;
	else brakecmd = MIN(1.0f, (float) (-(MAX(10.0, brakedelay * barkefactor))*x));
}

void Driver::calcSpeed()
{
	accelcmd = brakecmd = 0.0f;
	double speed;
	
	if (maxCarSpeed < racespeed) {
		maxCarSpeed = racespeed;
		speed = maxCarSpeed;
	} else speed = racespeed;

	if (Mode == mode_avoiding) {
		speed = avoidspeed;
	} else if (Mode == mode_correcting) {
		speed = racespeed - (racespeed - avoidspeed) * MAX(0.0, (correcttimer - simtime) / 7.0);
	}  
	minCarSpeed = speed;
	if (minCarSpeed > avoidspeed) {
		minCarSpeed = avoidspeed;
	}

	setAccelBrake(speed, 0.7);

	if (Mode == mode_pitting) {
		// Set pitEntry and pitExit car speed.
		getPitSpeed(speed); 
	}
/*
	setAccelBrake(speed, 0.8);
	if (Mode != mode_normal && drv_verbose) {
	  fprintf(stderr,"## maxSpeed=%.2f avoidSpeed=%.2f speed=%.2f X=%.2f\n", maxCarSpeed, avoidspeed, speed, x);
	}
*/
}


void Driver::getPitSpeed(double speed)
{
  //if (onPitArea() && Mode == mode_pitting) {
	double posPitEntry = pit->getPosEntry(); // beginning of pitEntry
	double posPitStart = pit->getLimitEntry(); // beginning of pitStart
	double posPitEnd = pit->getLimitExit(); // end of pitEnd
	double posPitExit = pit->getPosExit();  // end of pitExit
	double pitExitLength = posPitExit - posPitEnd;
	double pitLimitSpeed = pit->getPitstop() ? pitEntrySpeed : pitExitSpeed;
	 // distance before stopping the pitExit limited speed.
	double posExitMargin = pitExitLength - pitExitMargin;	// pitExitMargin variable in XML file (default = 80).
	if (posExitMargin < 0.0) {	// after the end of pitExit.
		posExitMargin = 0.0;
	} else if (posExitMargin > posPitExit) { // immediatly after pitEnd.
		posExitMargin = posPitExit;
	}
	//double pitExitDiff = pitExitLength - posExitMargin;
	/* TODO: (this pit exit doesn't works correctly!)
		Add a correct steering to exit of the Pit.
		Add border margin at the end of pitExit?
	*/

  // pitEntry Speed
	// pitEntry and pitStart are both before or after starting line.
	if (posPitEntry < posPitStart) {
  		if ( posFromStart >= posPitEntry && posFromStart <= posPitStart) {
			speed = pitEntrySpeed;
		}
	// pitEntry and pitStart are not on the same side of the starting line.
  	} else if ( posFromStart >= posPitEntry && posFromStart >= posPitStart) {
		speed = pitEntrySpeed;
	}
	
  // pitExit Speed
	// pitStart and pitEnd are on the same side of the starting line.
	if (posPitStart <= posPitEnd) {
			// pitExit on the same side too.
			if (posPitExit > posPitEnd && (posFromStart > posPitEnd && posFromStart <= posPitExit - posExitMargin)) {
				speed = pitLimitSpeed;
			// else after startlane
			} else if (posPitExit < posPitEnd && posFromStart <= posPitEnd) {
				posExitMargin = posPitExit -pitExitMargin;
				if (posFromStart <= posPitExit - posExitMargin) {
					speed = pitLimitSpeed;
				}
			}
	// startLane is on the middle of the pits
	} else if (posFromStart <= posPitExit) {
			if (posFromStart >= posPitEnd) {
				speed = pitLimitSpeed;		
			}
	}
		
		setAccelBrake(speed, 0.6);
 
  //}
}

void Driver::setControls()
{
	if (isStuck) {
	        stucksteer =  -mycardata->getCarAngle() / mCar->_steerLock;
		mCar->_steerCmd = -mycardata->getCarAngle() / mCar->_steerLock;
		mCar->_gearCmd = -1;		// Reverse gear.
		mCar->_clutchCmd = 0.0f;	// Full clutch (gearbox connected with engine).
		mCar->_brakeCmd = 0.0f;		// No brakes.
		mCar->_accelCmd = 1.0f;		// 100% accelerator pedal.
	} else {
#if 0
		// *FIX: Locus position at start lane* //
		if (mSitu->currentTime < 1)
		  mCar->_steerCmd = 0;
		else {
			mCar->_steerCmd = filterTrk(getSteer(mSitu));
		}
#endif
		mCar->_steerCmd = filterTrk(getSteer(mSitu));
		mCar->_gearCmd = getGear();
		//calcSpeed();
		mCar->_brakeCmd = filterABS(filterBrakeSpeed(filterBColl(filterBPit(getBrake()))));
		if (mCar->_brakeCmd == 0.0f) {
			mCar->_accelCmd = filterTCL(filterOverlap(getAccel()));
		} else {
			mCar->_accelCmd = 0.0f;
  		}
		mCar->_clutchCmd = getClutch();
	}
}


/**********************************************/
/* Compute Radius */
void Driver::computeRadius(int ref_line, float *radius)
{
	tTrackSeg *currentseg, *startseg = track->seg;
	currentseg = startseg;

	do {
		if (currentseg->type == TR_STR) 
		  {
		    radius[currentseg->id] = FLT_MAX;
		  } 
		else  /* turn */
		  {
		    radius[currentseg->id]  = currentseg->radius + currentseg->width/2.0;
		  }
		currentseg = currentseg->next;

	} while (currentseg != startseg);	

	/* override old radii with new */
	/* this needs major cleanup */

	currentseg = startseg;
	do 
	  {
	    //radius[currentseg->id] = line.lineRadius(ref_line, currentseg->lgfromstart);
	    radius[currentseg->id] = line.lineRadius(2, currentseg->lgfromstart);

	    if ( radius[currentseg->id] < 0 )
	      radius[currentseg->id] = FLT_MAX;

	    currentseg = currentseg->next;
	  } 
	while (currentseg != startseg);	
}

/***************************/
// Compute the length to the end of the segment.
float Driver::getDistToSegEnd()
{
	if (mCar->_trkPos.seg->type == TR_STR) {
		return mCar->_trkPos.seg->length - mCar->_trkPos.toStart;
	} else {
		return (mCar->_trkPos.seg->arc - mCar->_trkPos.toStart)*mCar->_trkPos.seg->radius;
	}
}

/***************************/
// Compute fitting acceleration.
float Driver::getAccel()
{
	if (mCar->_gear > 0) {
		accelcmd = MIN(1.0f, accelcmd);
		if (fabs(angle) > 0.8 && getSpeed() > 10.0f)
		  accelcmd = MAX(0.0f, MIN(accelcmd, 1.0f - getSpeed()/100.0f * fabs(angle)));
		return accelcmd;
	} else {
		return 1.0;
	}
}

/***************************/
// If we get lapped reduce accelerator.
float Driver::filterOverlap(float accel)
{
	int i;
	for (i = 0; i < opponents->getNOpponents(); i++) {
		if (opponent[i].getState() & OPP_LETPASS) {
		  
		  //return MIN(accel, 0.5f);
			return MIN(accel, LetPass);
		}
	}
	return accel;
}

/***************************/
// Compute initial brake value.
float Driver::getBrake()
{
	// Car drives backward?
	if (mCar->_speed_x < -MAX_UNSTUCK_SPEED) {
		// Yes, brake.
		return 1.0;
	} else {
		// We drive forward, normal braking.
		return brakecmd;
	}
}

/***************************/
// Compute gear.
int Driver::getGear()
{
  if (mCar->_gear <= 0) {
	return 1;
  }

  if ( CHANGE_GEAR ) {
	/* BT gear changing */
	float gr_up = mCar->_gearRatio[mCar->_gear + mCar->_gearOffset];
	float omega = mCar->_enginerpmRedLine/gr_up;
	float wr = mCar->_wheelRadius(2);

	if (omega*wr*SHIFT < mCar->_speed_x) {
		//return mCar->_gear + 1;
		return (mCar->_gear + 1 > MAX_GEAR ? MAX_GEAR : mCar->_gear + 1);
	} else {
		float gr_down = mCar->_gearRatio[mCar->_gear + mCar->_gearOffset - 1];
		omega = mCar->_enginerpmRedLine/gr_down;
		if (mCar->_gear > 1 && omega*wr*SHIFT > mCar->_speed_x + SHIFT_MARGIN) {
			return mCar->_gear - 1;
		}
	}
  } else {
	  // Maybe you need to udate SHIFT_UP / SHIFT_DOWN for your Car?
   	  //  by default SHIFT_DOWN = 120  SHIFT_UP = 0.99
  	if (mCar->_enginerpm / mCar->_enginerpmRedLine > SHIFT_UP) 
   	{
    		return mCar->_gear + 1;
   	} else {
    		float gratio_dwn = mCar->_gearRatio[mCar->_gear + mCar->_gearOffset - 1] / mCar->_gearRatio[mCar->_gear + mCar->_gearOffset];
    		if (mCar->_gear > 1 && (mCar->_enginerpmRedLine - SHIFT_DOWN) / mCar->_enginerpm > gratio_dwn) 
    		{
      		return mCar->_gear - 1;
    		}
  	}
  }
	return mCar->_gear;
}
/***************************/
// Compute the clutch value.
float Driver::getClutch()
{
 float speedr;
 float omega;
 float wr = mCar->_wheelRadius(2);
 //int PrevGear = mCar->_gear - 1;
 omega = mCar->_enginerpmRedLine/mCar->_gearRatio[mCar->_gear + mCar->_gearOffset];
 speedr = (CLUTCH_SPEED + MAX(0.0f, mCar->_speed_x))/fabs(wr*omega);
 //double AngleToTrack = RtTrackSideTgAngleL(&(mCar->_trkPos)) - mCar->_yaw;
 
#if 0
	/* BT Cluth */
	if (mCar->_gear > 1) 
	{
		clutchtime = 0.0f;
		return 0.0f;
	} else {
		float drpm = mCar->_enginerpm - mCar->_enginerpmRedLine/2.0f;
		clutchtime = MIN(CLUTCH_FULL_MAX_TIME, clutchtime);
		float clutcht = (CLUTCH_FULL_MAX_TIME - clutchtime)/CLUTCH_FULL_MAX_TIME;
		if (mCar->_gear == 1 && mCar->_accelCmd > 0.0f) {
			clutchtime += (float) RCM_MAX_DT_ROBOTS;
		}

		if (drpm > 0) {
			//float speedr;
			if (mCar->_gearCmd == 1) {
				// Compute corresponding speed to engine rpm.
				//float omega = mCar->_enginerpmRedLine/mCar->_gearRatio[mCar->_gear + mCar->_gearOffset];
				//float wr = mCar->_wheelRadius(2);
				//speedr = (CLUTCH_SPEED + MAX(0.0f, mCar->_speed_x))/fabs(wr*omega);
				float clutchr = MAX(0.0f, (1.0f - speedr*2.0f*drpm/mCar->_enginerpmRedLine));
				return MIN(clutcht, clutchr);
			} else {
				// For the reverse gear.
				clutchtime = 0.0f;
				return 0.0f;
			}
		} else {
			return clutcht;
		}
	}
#endif

	if (1 || mCar->_gearCmd > 1) 
	{
		float maxtime = MAX(0.06f, 0.32f - ((float) mCar->_gearCmd / 65.0f));
		if (mCar->_gear != mCar->_gearCmd)
			clutchtime = maxtime;
		if (clutchtime > 0.0f)
			clutchtime -= (float) (RCM_MAX_DT_ROBOTS * (0.02f + ((float) mCar->_gearCmd / 8.0f)));
		return 2.0f * clutchtime;
	} else {
		float drpm = mCar->_enginerpm - mCar->_enginerpmRedLine/2.0f;
		float ctlimit = 0.9f;
		if (mCar->_gearCmd > 1)
			ctlimit -= 0.15f + (float) mCar->_gearCmd/13.0f;
		clutchtime = MIN(ctlimit, clutchtime);
		if (mCar->_gear != mCar->_gearCmd)
			clutchtime = 0.0f;
		float clutcht = (ctlimit - clutchtime) / ctlimit;
		if (mCar->_gear == 1 && mCar->_accelCmd > 0.0f) 
		{
			clutchtime += (float) RCM_MAX_DT_ROBOTS;
		}

		if (mCar->_gearCmd == 1 || drpm > 0) 
		{
			
			if (mCar->_gearCmd == 1) 
			{
				// Compute corresponding speed to engine rpm.
				//float omega = mCar->_enginerpmRedLine/mCar->_gearRatio[mCar->_gear + mCar->_gearOffset];
				//float wr = mCar->_wheelRadius(2);
				//speedr = (CLUTCH_SPEED + MAX(0.0f, mCar->_speed_x))/fabs(wr*omega);
				float clutchr = MAX(0.0f, (1.0f - speedr*2.0f*drpm/mCar->_enginerpmRedLine)) * 
					(mCar->_gearCmd == 1 ? 0.95f : (0.7f - (float)(mCar->_gearCmd)/30.0f));
				return MIN(clutcht, clutchr);
			} else {
				// For the reverse gear.
				clutchtime = 0.0f;
				return 0.0f;
			}
		} else {
			return clutcht;
		}
	}

}

/***************************/
void Driver::setMode( int newmode )
{
	if (Mode == newmode)
		return;

	if (Mode == mode_normal || Mode == mode_pitting)
	{
		correcttimer = simtime + 7.0;
		correctlimit = 1000.0;
	}

	Mode = newmode;
}

/***************************/
// Compute steer value.
float Driver::getSteer(tSituation *s)
{
	float targetAngle, racesteer;

	 /* K1999 raceline */
	 raceline.GetRaceLineData( mSitu, &racetarget, &racespeed, &avoidspeed, &raceoffset, &rlookahead, &racesteer );
	 
	vec2f target = getTargetPoint();
	float avoidsteer = 0.0f;
	float steer = 0.0f;


	{
		targetAngle = atan2(target.y - mCar->_pos_Y, target.x - mCar->_pos_X);
		avoidsteer = calcSteer( targetAngle, 0, racesteer );
	}

	if (Mode == mode_pitting)
	{
		return avoidsteer;
	}

        /* K1999 steering */
	targetAngle = atan2(racetarget.y - mCar->_pos_Y, racetarget.x - mCar->_pos_X);

	NORM_PI_PI(targetAngle);

	if (Mode == mode_avoiding && 
			(!avoidMode ||
			 (drvmode == avoidright && raceoffset >= myoffset && raceoffset < avoidlftoffset) ||
			 (drvmode == avoidleft && raceoffset <= myoffset && raceoffset > avoidrgtoffset)))
	{
		// we're avoiding, but trying to steer somewhere the raceline takes us.
		// hence we'll just correct towards the raceline instead.
		setMode( mode_correcting );
		if (mode_verbose)
			fprintf(stderr,"%s enter in correcting mode\n", mCar->_name);
	}


	/* K1999 raceline */
	  int line;
	  line = LINE_RL;

	if (Mode == mode_correcting && (lastmode == mode_normal ||
			 (fabs(angle) < 0.2f && fabs(racesteer) < 0.4f && fabs(laststeer - racesteer) < 0.05 &&
			((fabs(mCar->_trkPos.toMiddle) < mCar->_trkPos.seg->width/2 - 1.0) || mCar->_speed_x < 10.0) &&
			(raceline.isOnLine(line)))))
	{
		// we're correcting & are now close enough to the raceline to
		// switch back to 'normal' mode...
		setMode( mode_normal );
		if (mode_verbose)
			fprintf(stderr,"%s enter in Normal mode\n", mCar->_name);
	}

	if (Mode == mode_normal)
	{
		steer = racesteer;
		lastNSasteer = racesteer*0.8;
	}
	else
	{
		if (Mode != mode_correcting)
		{
			correctlimit = 1000.0;
			correcttimer = simtime + 7.0;
			steer = smoothSteering( avoidsteer );
		}
		else
			steer = smoothSteering(correctSteering( avoidsteer, racesteer ));
	
		if (fabs(angle) >= 1.6)
		{
			if (steer > 0.0)
				steer = 1.0;
			else
				steer = -1.0;
		}
	}

	return steer;
}

/***************************/
float Driver::calcSteer( float targetAngle, int rl, float racesteer )
{
	double steer_direction = targetAngle - mCar->_yaw;

	NORM_PI_PI(steer_direction);

	float steer = (float) (steer_direction/mCar->_steerLock);

	// smooth steering.  I know there's a separate function for this, but what the hey!
	if (Mode != mode_pitting)
	{
	        //double minspeedfactor = (((60.0 - (MAX(40.0, MIN(70.0, getSpeed() + MAX(0.0, mCar->_accel_x*5))) - 25)) / 300) * MAX(10.0, 30.0+mCar->_accel_x));
		double minspeedfactor = (((80.0 - (MAX(40.0, MIN(70.0, getSpeed() + MAX(0.0, mCar->_accel_x*5))) - 25)) / 300) * (5.0 + MAX(0.0, (CA-1.9)*20)));

		double maxspeedfactor = minspeedfactor;
		double rInverse = raceline.getRInverse();

		if (rInverse > 0.0)
		{
		        //minspeedfactor = MAX(minspeedfactor/3, minspeedfactor - rInverse*getSpeed() + MAX(0.0f, angle/100));
			minspeedfactor = MAX(minspeedfactor/3, minspeedfactor - rInverse*80.0);//getSpeed() + MAX(0.0f, angle/100));
			maxspeedfactor = MAX(maxspeedfactor/3, maxspeedfactor + rInverse*20.0);//getSpeed() + MIN(0.0f, angle/100));
		}
		else
		{
		        //maxspeedfactor = MAX(maxspeedfactor/3, maxspeedfactor + rInverse*getSpeed() + MIN(0.0f, angle/100));
			maxspeedfactor = MAX(maxspeedfactor/3, maxspeedfactor + rInverse*80.0);//getSpeed() + MIN(0.0f, angle/100));
			minspeedfactor = MAX(minspeedfactor/3, minspeedfactor + rInverse*20.0);//getSpeed() + MAX(0.0f, angle/100));
		}
			
		steer = (float) MAX(lastNSasteer - minspeedfactor, MIN(lastNSasteer + maxspeedfactor, steer));
	}

	lastNSasteer = steer;

	if (fabs(angle) > fabs(speedangle))
	{
		// steer into the skid
		double sa = MAX(-0.3, MIN(0.3, speedangle/3));
		//double anglediff = (sa - angle) * 0.7;
		double anglediff = (sa - angle) * (0.7 - MAX(0.0, MIN(0.3, mCar->_accel_x/100)));
		//anglediff += raceline.getRInverse() * 10;
		steer += (float) (anglediff*0.7);
	}
	
	if (fabs(angle) > 1.2)
	{
		if (steer > 0.0f)
			steer = 1.0f;
		else
			steer = -1.0f;
	}
	else if (fabs(mCar->_trkPos.toMiddle) - mCar->_trkPos.seg->width/2 > 2.0)
	{
		steer = (float) MIN(1.0f, MAX(-1.0f, steer * (1.0f + (fabs(mCar->_trkPos.toMiddle) - mCar->_trkPos.seg->width/2)/14 + fabs(angle)/2)));
	}

	
	if (Mode != mode_pitting)
   	{
	   // limit how far we can steer against raceline 
	   double limit = (90.0 - MAX(40.0, MIN(60.0, mCar->_speed_x))) / (50 + fabs(angle)*fabs(angle)*3);
	   steer = MAX(racesteer-limit, MIN(racesteer+limit, steer));
   	}
	
	return steer;
}

/***************************/
float Driver::correctSteering( float avoidsteer, float racesteer )
{
	/* K1999 raceline */
	  int line;
	  line = LINE_RL;
	  
	float steer = avoidsteer;

	double speed = MAX(50.0, getSpeed());
	//double changelimit = MIN(1.0, raceline.correctLimit(line));
	double changelimit = MIN(raceline.correctLimit(line), (((120.0-getSpeed())/6000) * (0.5 + MIN(fabs(avoidsteer),fabs(racesteer))/10)));

	if (Mode == mode_correcting && simtime > 2.0f)
	{
		// move steering towards racesteer...
		if (correctlimit < 900.0)
		{
			if (steer < racesteer)
			{
				if (correctlimit >= 0.0)
				{
				        //steer = (float) MIN(racesteer, steer + correctlimit);
				        steer = racesteer;
				}
				else
				{
					steer = (float) MIN(racesteer, MAX(steer, racesteer + correctlimit));
				}
			}
			else
			{
				if (correctlimit <= 0.0)
				{
					//steer = (float) MAX(racesteer, steer-correctlimit);
				        steer = racesteer;
				}
				else
				{
					steer = (float) MAX(racesteer, MIN(steer, racesteer + correctlimit));
				}
			}
		}

		speed -= mCar->_accel_x/10;
		speed = MAX(55.0, MIN(150.0, speed + (speed*speed/55)));
		double rInverse = raceline.getRInverse() * (mCar->_accel_x<0.0 ? 1.0 + fabs(mCar->_accel_x)/10.0 : 1.0);
		double correctspeed = 0.5;
		if ((rInverse > 0.0 && racesteer > steer) || (rInverse < 0.0 && racesteer < steer))
			correctspeed += rInverse*110;

 		if (racesteer > steer)
			steer = (float) MIN(racesteer, steer + changelimit);
 		else
			steer = (float) MAX(racesteer, steer - changelimit);

 		//correctlimit = (steer - racesteer) * 1.08;
 		correctlimit = (steer - racesteer);
	}

	return steer;
}

/***************************/
float Driver::smoothSteering( float steercmd )
{
	/* try to limit sudden changes in steering to avoid loss of control through oversteer. */
	
	//double speedfactor = (((60.0 - (MAX(40.0, MIN(70.0, getSpeed() + MAX(0.0, mCar->_accel_x*5))) - 25)) / 300) * 1.5) / 0.785;
	//double rearskid = MAX(0.0, MAX(mCar->_skid[2], mCar->_skid[3]) - MAX(mCar->_skid[0], mCar->_skid[1]));
	double speedfactor = (((60.0 - (MAX(40.0, MIN(70.0, getSpeed() + MAX(0.0, mCar->_accel_x*5))) - 25)) / 300) * 1.2) / 0.785;
	if (fabs(steercmd) < fabs(laststeer) && fabs(steercmd) <= fabs(laststeer - steercmd))
		speedfactor *= 2;
	steercmd = (float) MAX(laststeer - speedfactor, MIN(laststeer + speedfactor, steercmd));
	return steercmd;
}

/***************************/
// Compute target point for steering.
vec2f Driver::getTargetPoint()
{
	tTrackSeg *seg = mCar->_trkPos.seg;
	float lookahead;
	getMyOffset();
	getOppState();
	oppLetPass();
	float offset = myoffset;
	float length = getDistToSegEnd();
	float pitoffset = -100.0f;

	if ( drv_verbose)
		GfOut("# %s: Offset = %f\n", mCar->_name, offset);

	if (pit->getInPit()) {
		// To stop in the pit we need special lookahead values.
		if (currentspeedsqr > pit->getSpeedlimitSqr()) {
			lookahead = PIT_LOOKAHEAD + mCar->_speed_x*LOOKAHEAD_FACTOR;
		} else {
			lookahead = PIT_LOOKAHEAD;
		}
	} else {
		// Usual lookahead.
		lookahead = rlookahead;

		double speed = MAX(20.0, getSpeed());// + MAX(0.0, mCar->_accel_x));

		//lookahead = LOOKAHEAD_CONST + mCar->_speed_x*LOOKAHEAD_FACTOR;
		//lookahead = MAX(lookahead, LOOKAHEAD_CONST + ((mCar->_speed_x*(mCar->_speed_x/2)) / 60.0));

		lookahead = (float) (LOOKAHEAD_CONST * 1.2 + speed * 0.60);
		lookahead = MIN(lookahead, (float) (LOOKAHEAD_CONST + ((speed*(speed/7)) * 0.15)));

		// Prevent "snap back" of lookahead on harsh braking.
		float cmplookahead = lastlookahead - mCar->_speed_x*RCM_MAX_DT_ROBOTS;
		if (lookahead < cmplookahead) {
			lookahead = cmplookahead;
		}
	}

	lastlookahead = lookahead;

	// Search for the segment containing the target point.
	while (length < lookahead) {
		seg = seg->next;
		length += seg->length;
	}
	// Length now > lookahead

	length = lookahead - length + seg->length;
	// length now distance past start of seg
	
	float fromstart = seg->lgfromstart;
	fromstart += length;

	// Compute the target point.
	pitoffset = pit->getPitOffset(pitoffset, fromstart);
	if ((pit->getPitstop() || pit->getInPit()) && pitoffset != -100.0f)
	{
		setMode(mode_pitting);
		offset = myoffset = pitoffset;
		if (mode_verbose)
			fprintf(stderr,"%s enter in pitting mode\n", mCar->_name);
	}
	else if (Mode == mode_pitting)
	     {
		setMode(mode_correcting);
		if (mode_verbose)
			fprintf(stderr,"%s enter in correcting mode\n", mCar->_name);
	     }

	vec2f s;
	if (Mode != mode_pitting)
	{
		
		/* K1999 raceline */
	        raceline.GetPoint( offset, lookahead, &s );
  	        return s;
	}

	s.x = (seg->vertex[TR_SL].x + seg->vertex[TR_SR].x)/2.0f;
	s.y = (seg->vertex[TR_SL].y + seg->vertex[TR_SR].y)/2.0f;

	if ( seg->type == TR_STR) {
		vec2f d, n;
		n.x = (seg->vertex[TR_EL].x - seg->vertex[TR_ER].x)/seg->length;
		n.y = (seg->vertex[TR_EL].y - seg->vertex[TR_ER].y)/seg->length;
		n.normalize();
		d.x = (seg->vertex[TR_EL].x - seg->vertex[TR_SL].x)/seg->length;
		d.y = (seg->vertex[TR_EL].y - seg->vertex[TR_SL].y)/seg->length;
		return s + d*length + offset*n;
	} else {
		vec2f c, n, t, rt;
		c.x = seg->center.x;
		c.y = seg->center.y;
		float arc = length/seg->radius;
		float arcsign = (seg->type == TR_RGT) ? -1.0f : 1.0f;
		arc = arc*arcsign;
		s = s.rotate(c, arc);

		n = c - s;
		n.normalize();
		t = s + arcsign*offset*n;

		if (Mode != mode_pitting)
		{
		  
			/* K1999 raceline */
		        raceline.GetPoint( offset, lookahead, &rt );
		   
			double dx = t.x - mCar->_pos_X;
			double dy = t.y - mCar->_pos_Y;
			double dist1 = sqrt(dx*dx + dy*dy);
			dx = rt.x - mCar->_pos_X;
			dy = rt.y - mCar->_pos_Y;
			double dist2 = sqrt(dx*dx + dy*dy);
			if (dist2 > dist1)
				t = rt;
		}

		return t;
	}
}

/***************************/
void Driver::getMyOffset()
{
	avoidMode = false;
	drvmode = 0;
	myoffset = mCar->_trkPos.toMiddle;
	//carBorderdist = mCar->_trkPos.seg->width / 2.0 - fabs(myoffset) - mCar->_dimension_y / 2.0;
	double borderDist = track->width/2 - mCar->_dimension_y;
	avoidlftoffset = MAX(myoffset, mCar->_trkPos.seg->width / 2 - 1.5);
	avoidrgtoffset = MIN(myoffset, -(mCar->_trkPos.seg->width / 2 - 1.5));
 	leftOffset = borderDist; // limit to the left
	rightOffset = -borderDist; // limit to the right
	rInverse = raceline.getRInverse();
	int offlft = 0;
	int offrgt = 0;

	/* Increment speed dependent. */
	double incspeed = MIN(60.0, MAX(45.0, getSpeed())) - 20.0;
	//incfactor = (MAX_INC_FACTOR - MIN(fabs(mCar->_speed_x)/MAX_INC_FACTOR, (MAX_INC_FACTOR - 1.0f))) * 4.0f;
	incfactor = (MAX_INC_FACTOR - MIN(fabs(incspeed)/MAX_INC_FACTOR, (MAX_INC_FACTOR - 1.0f))) * (12.0f + MAX(0.0, (CA-1.9)*14));
	rgtinc = incfactor * MIN(1.3, MAX(0.4, 1.0 + rInverse * (rInverse<0.0?20:80)));
	lftinc = incfactor * MIN(1.3, MAX(0.4, 1.0 - rInverse * (rInverse>0.0?20:80)));


	offlft = (myoffset > mCar->_trkPos.seg->width/2 - 1.5);
	offrgt = (myoffset < -(mCar->_trkPos.seg->width/2 - 1.5));
	if (offlft) {
		myoffset -= OVERTAKE_OFFSET_INC*rgtinc/2;
	} else if (offrgt) {
		myoffset += OVERTAKE_OFFSET_INC*lftinc/2;
	}
	avoidlftoffset = MAX(avoidlftoffset, myoffset - OVERTAKE_OFFSET_INC*rgtinc*(offlft ? 6 : 2));
	avoidrgtoffset = MIN(avoidrgtoffset, myoffset + OVERTAKE_OFFSET_INC*lftinc*(offrgt ? 6 : 2));
	

	if (myoffset < rightOffset) // we're already outside right limit, bring us back towards track
	{
		rightOffset = myoffset + OVERTAKE_OFFSET_INC*lftinc;
		leftOffset = MIN(leftOffset, myoffset+OVERTAKE_OFFSET_INC*lftinc*2);
	}
	else if (myoffset > leftOffset) // outside left limit, bring us back
	{
		leftOffset = myoffset - OVERTAKE_OFFSET_INC*rgtinc;
		rightOffset = MAX(rightOffset, myoffset-OVERTAKE_OFFSET_INC*rgtinc*2);
	}
	else
	{
		// set tighter limits based on how far we're allowed to move 
		leftOffset = MIN(leftOffset, myoffset+OVERTAKE_OFFSET_INC*lftinc*2);
		rightOffset = MAX(rightOffset, myoffset-OVERTAKE_OFFSET_INC*rgtinc*2);
	}

}
/***************************/
void Driver::getOppState()
{
	// Check collision
	oppCollision();

	Opponent *oppnt = NULL;
	float catchdist, mincatchdist = 500.0f;
	mincatchdist = MAX(30.0, 1500.0 - fabs(raceline.getRInverse()) * 10000);
	int otry;
	int i;
	int otry_success = 0;
 
	// try to ovetake opponent
	for (otry=0; otry<=1; otry++)
	{
	  for (i = 0; i < opponents->getNOpponents(); i++) 
	  {
		tCarElt *oppCar = opponent[i].getCarPtr();
 
		if (!(oppCar->_state & (RM_CAR_STATE_NO_SIMU & ~RM_CAR_STATE_PIT)) && !(opponent[i].getState() & OPP_IGNORE)) {
			if (opponent[i].getState() & OPP_FRONT_FOLLOW) {
				continue;
			} else if ((opponent[i].getState() & OPP_FRONT) &&
			    	!(opponent[i].isTeamMate() && mCar->race.laps <= opponent[i].getCarPtr()->race.laps))
 			{
				if ( opp_verbose )
					GfOut("#%s check Opponent %s\n",mCar->_name, oppCar->_name);

				double otry_factor = (otry ? (0.2 + (1.0 - ((currentsimtime-avoidtime)/7.0)) * 0.8) : 1.0);
				double distance = opponent[i].getDistance() * otry_factor;
				double mySpeed = getSpeed();
				double mspeed = MIN(avoidspeed, mySpeed + MAX(0.0, (10.0 - distance)));				
				double ospeed = opponent[i].getSpeed();
				catchdist = (float) MIN(mspeed*distance/(mspeed-ospeed), distance*CATCH_FACTOR) * otry_factor;
				double oppMinDist = MIN(fabs(mspeed-ospeed)*2, mCar->_dimension_x*3);

				//if (ospeed <= mySpeed && catchdist < mincatchdist && distance <= mCar->_dimension_x*4) {
				if (catchdist < mincatchdist && distance < oppMinDist) {
					mincatchdist = catchdist;
					oppnt = &opponent[i];
					otry_success = otry;
 				}
				
 			}
			continue;
		}
 	  }
		if (oppnt) break;
		if (Mode != mode_avoiding) break;
 	}

	if (oppnt != NULL) {
		if (mode_verbose)
			fprintf(stderr,"%s enter in Avoiding mode\n", mCar->_name);
		setMode( mode_avoiding );
		if (otry_success == 0)
			avoidtime = currentsimtime;
 		tCarElt *oppCar = oppnt->getCarPtr();

		/* Compute the opponents distance to the middle.*/
		float oppCarTM = oppCar->_trkPos.toMiddle;
		float sidemargin = oppnt->getWidth() + getWidth() + 2.0f;
		float sidedist = fabs(oppCar->_trkPos.toLeft - mCar->_trkPos.toLeft);

		if ((oppCarTM < -(oppCar->_trkPos.seg->width - 5.0) && rInverse < 0.0) ||
		    (oppCarTM > (oppCar->_trkPos.seg->width - 5.0) && rInverse > 0.0))
		{
			sidemargin += fabs(rInverse) * 150;
		}

		if (oppCarTM > (oppCar->_trkPos.seg->width - 5.0) ||
		    (mCar->_trkPos.toLeft > oppCar->_trkPos.toLeft && (sidedist < sidemargin || (oppnt->getState() & OPP_COLL))))
		{
			myoffset -= OVERTAKE_OFFSET_INC*rgtinc;
			drvmode = avoidleft;
		} 
		else if (oppCarTM < -(oppCar->_trkPos.seg->width - 5.0) ||
		         (mCar->_trkPos.toLeft < oppCar->_trkPos.toLeft && (sidedist < sidemargin || (oppnt->getState() & OPP_COLL))))
		{
			myoffset += OVERTAKE_OFFSET_INC*lftinc;
			drvmode = avoidright;
		} 

		else 
		{
			/* If the opponent is near the middle 
			   we try to move the offset toward the inside of the expected turn.
			   Try to find out the characteristic of the track up to catchdist. */
			tTrackSeg *seg = mCar->_trkPos.seg;
			float length = getDistToSegEnd();
			float oldlen, seglen = length;
			float lenright = 0.0f, lenleft = 0.0f;
			mincatchdist = MIN(mincatchdist, DISTCUTOFF);

			do {
				switch (seg->type) {
				case TR_LFT:
					lenleft += seglen;
					break;
				case TR_RGT:
					lenright += seglen;
					break;
				default:
					// Do nothing.
					break;
				}
				seg = seg->next;
				seglen = seg->length;
				oldlen = length;
				length += seglen;
			} while (oldlen < mincatchdist);

			/* If we are on a straight look for the next turn. */
			if (lenleft == 0.0f && lenright == 0.0f) {
				while (seg->type == TR_STR) {
					seg = seg->next;
				}
				/* Assume: left or right if not straight. */
				if (seg->type == TR_LFT) {
					lenleft = 1.0f;
				} else {
					lenright = 1.0f;
				}
			}

			/* Because we are inside we can go to the border. */
			//float maxoff = (float) ((oppCar->_trkPos.seg->width - mCar->_dimension_y)/2.0f-BORDER_OVERTAKE_MARGIN*2);

			if ((lenleft>lenright && rInverse < 0.0) ||
			    (lenleft<=lenright && rInverse > 0.0))
			{
				/* avoid more if on the outside of opponent on a bend. 
				  Stops us from cutting in too much and colliding...*/
				sidemargin += fabs(rInverse) * 150;
			}

			if (lenleft > lenright) {
			        if (sidedist<sidemargin || (oppnt->getState() & OPP_COLL)) {
					myoffset += OVERTAKE_OFFSET_INC*lftinc*0.7;
 				}
				drvmode = avoidright;
				avoidMode = true;
			} else {
				if (sidedist<sidemargin || (oppnt->getState() & OPP_COLL)) {
					myoffset -= OVERTAKE_OFFSET_INC*rgtinc*0.7;
 				}
				drvmode = avoidleft;
				avoidMode = true;
			}
		}

		myoffset = MIN(avoidlftoffset, MAX(avoidrgtoffset, myoffset));
		myoffset = MIN(leftOffset, MAX(rightOffset, myoffset));
	} 
}

/***************************/
void Driver::oppCollision()
{
  Opponent *oppnt = NULL;
  int i;

	for (i = 0; i < opponents->getNOpponents(); i++) 
	{
		tCarElt *oppCar = opponent[i].getCarPtr();

		/* Ignore Pitting Cars */
		if (oppCar->_state > RM_CAR_STATE_PIT)
			continue;

		/* Ignore offtrack cars */
		if ( (fabs(oppCar->_trkPos.toMiddle) > mCar->_trkPos.seg->width/2 + 3.0) && 
		     (fabs(oppCar->_trkPos.toMiddle - mCar->_trkPos.toMiddle) >= 5.0)      )
			continue;

		if ((opponent[i].getState() & OPP_SIDE) || (opponent[i].getState() & OPP_LETPASS))
		{
			if (mode_verbose)
				fprintf(stderr,"%s enter in Avoiding mode\n", mCar->_name);
			setMode( mode_avoiding );
			oppnt = &opponent[i];
			
			float sidedist = fabs(oppCar->_trkPos.toLeft - mCar->_trkPos.toLeft);
			float sidemargin = opponent[i].getWidth() + getWidth() + 2.0f;
			float side = mCar->_trkPos.toMiddle - oppCar->_trkPos.toMiddle;

			if ((side > 0.0 && rInverse < 0.0) ||
			    (side < 0.0 && rInverse > 0.0))
			{
				/* avoid more if on the outside of opponent on a bend.
				  Stops us from cutting in too much and colliding...*/
				sidemargin += fabs(rInverse) * 150;
			}

			double sidedist2 = sidedist;
			if (side > 0.0)
			{
				sidedist2 -= (oppnt->getSpeedAngle() - speedangle) * 40;
				sidemargin -= MIN(0.0, rInverse*100);
			}
			else
			{
				sidedist2 -= (speedangle - oppnt->getSpeedAngle()) * 40;
				sidemargin += MAX(0.0, rInverse*100);
			}
			sidedist = MIN(sidedist, sidemargin);

			if (sidedist < sidemargin)
			{
			        //float side = mCar->_trkPos.toMiddle - oppCar->_trkPos.toMiddle;
			        //float w = mCar->_trkPos.seg->width/WIDTHDIV-BORDER_OVERTAKE_MARGIN;
			        double sdiff = (3.0 - (sidemargin-sidedist)/sidemargin);

				if (side > 0.0f) { 
					myoffset += (float) (OVERTAKE_OFFSET_INC*lftinc*MAX(0.2f, MIN(1.0f, sdiff)));
				} else {
				        myoffset -= (float) (OVERTAKE_OFFSET_INC*rgtinc*MAX(0.2f, MIN(1.0f, sdiff)));
				}
			}
			else if (sidedist > sidemargin + 3.0 && raceoffset > myoffset + OVERTAKE_OFFSET_INC*incfactor)
			  {
			        //myoffset += OVERTAKE_OFFSET_INC*incfactor;
				myoffset += OVERTAKE_OFFSET_INC*lftinc/4;
			  }
			else if (sidedist > sidemargin + 3.0 && raceoffset < myoffset - OVERTAKE_OFFSET_INC*incfactor)
			  {
			        //myoffset -= OVERTAKE_OFFSET_INC*incfactor;
				myoffset -= OVERTAKE_OFFSET_INC*rgtinc/4;
			  }


			if (oppCar->_trkPos.toLeft > mCar->_trkPos.toLeft)
			{
			        //avoidrgtoffset = (float) MAX(avoidrgtoffset, oppCar->_trkPos.toMiddle + (o->getWidth()+1.0f));
				drvmode = avoidright;
			}
			else
			{
			        //avoidlftoffset = (float) MIN(avoidlftoffset, oppCar->_trkPos.toMiddle - (o->getWidth()+1.0f));
				drvmode = avoidleft;
			}
			drvmode = avoidside;

			if (drvmode > 0) avoidMode = true;

			myoffset = MIN(avoidlftoffset, MAX(avoidrgtoffset, myoffset));
			myoffset = MIN(leftOffset, MAX(rightOffset, myoffset));
		}
	}
}

/***************************/
void Driver::oppLetPass()
{
	Opponent *oppnt = NULL;
	//oppnt = NULL;
	int i;
	bool TM_LETPASS = false;
	float mindist = -1000.0f;

	for (i = 0; i < opponents->getNOpponents(); i++)
	{
		/* just check temmate situ for debug and compute distance */
	    if ( opp_verbose ) {
		Opponent *opp = &opponent[i];
		if(opponent[i].isTeamMate()) {
		  if (opp->getCarPtr()->race.laps > 1) {
			if(opp->getDistance() > 0.1f && opp->getDistance()<= mCar->_dimension_x) {
			  	GfOut(" %s >> TeamMate:%s inFront\n", mCar->_name, opponent[i].getCarPtr()->_name);
			} else if(opp->getDistance() < 0.0f && opp->getDistance()>= -mCar->_dimension_x*2.0) {
            			GfOut("TEAMMATE %s, laps: %i olaps: %i diff_dam: %i dist: %.3f %.3f\n", 
					opponent[i].getCarPtr()->_name, mCar->race.laps, opp->getCarPtr()->race.laps, 
					mCar->_dammage - opp->getDamage(), opp->getDistance(), TEAM_REAR_DIST);
 			}
 		  }
		} 
	    }
		
 		// Set LetPass TeamMate (change Lead)
		if ( opponent[i].isTeamMate() &&
			mCar->race.laps <= opponent[i].getCarPtr()->race.laps &&
			(opponent[i].getDistance() > -TEAM_REAR_DIST && opponent[i].getDistance() <= -mCar->_dimension_x) &&
 			mCar->_dammage - opponent[i].getDamage() > MAX_DAMAGE_DIFF ) {
		  	 
			TM_LETPASS = true;
			if ( opp_verbose )
				GfOut("#%s LetPass TeamMate\n", mCar->_name);
		}

		if (((opponent[i].getState() & OPP_LETPASS) && !opponent[i].isTeamMate()) || TM_LETPASS) {
			// Behind, larger distances are smaller ("more negative").
			if (opponent[i].getDistance() > mindist) {
				mindist = opponent[i].getDistance();
				oppnt = &opponent[i];
			}
		}
	}

	if (oppnt != NULL) {
		tCarElt *oppCar = oppnt->getCarPtr();
		float side = mCar->_trkPos.toMiddle - oppCar->_trkPos.toMiddle;
		float w = mCar->_trkPos.seg->width/WIDTHDIV-BORDER_OVERTAKE_MARGIN;
		if (side > 0.0f) {
			if (myoffset < w) {
				myoffset += OVERTAKE_OFFSET_INC*lftinc/2;
			}
		} else {
			if (myoffset > -w) {
				myoffset -= OVERTAKE_OFFSET_INC*rgtinc/2;
			}
		}
		setMode( mode_avoiding );
		if (mode_verbose)
			fprintf(stderr,"%s enter in Avoiding mode\n", mCar->_name);

		myoffset = MIN(avoidlftoffset, MAX(avoidrgtoffset, myoffset));
		myoffset = MIN(leftOffset, MAX(rightOffset, myoffset));
		return;
	} 

#if 1       
	// no-one to avoid, work back towards raceline
	//if (Mode != mode_normal && fabs(myoffset) > mCar->_trkPos.seg->width/2 + 0.5)
	if (Mode != mode_normal && fabs(myoffset-raceoffset) > 1.0)
	{
		if (myoffset > raceoffset + OVERTAKE_OFFSET_INC * rgtinc/4)
			myoffset -= OVERTAKE_OFFSET_INC * rgtinc/4;
		else if (myoffset < raceoffset + OVERTAKE_OFFSET_INC * lftinc/4)
			myoffset += OVERTAKE_OFFSET_INC * lftinc/4;
	}
#endif
	if (simtime > 2.0f)
	{
		if (myoffset > raceoffset)
			myoffset = MAX(raceoffset, myoffset - OVERTAKE_OFFSET_INC*incfactor/2);
		else
			myoffset = MIN(raceoffset, myoffset + OVERTAKE_OFFSET_INC*incfactor/2);
	}

	myoffset = MIN(leftOffset, MAX(rightOffset, myoffset));
}

/***************************/
int Driver::isAlone()
{
	int i;
	for (i = 0; i < opponents->getNOpponents(); i++) {
		if (opponent[i].getState() & (OPP_COLL | OPP_LETPASS)) {
			return 0;	// Not alone.
		}
	}
	return 1;	// Alone.
}


/***************************/
// Compute aerodynamic downforce coefficient CA.
void Driver::initCa()
{
  char *WheelSect[4] = {(char *)SECT_FRNTRGTWHEEL, (char *)SECT_FRNTLFTWHEEL, (char *)SECT_REARRGTWHEEL, (char *)SECT_REARLFTWHEEL};
  double rearwingarea = GfParmGetNum(mCar->_carHandle, (char *)SECT_REARWING, (char *)PRM_WINGAREA, (char*) NULL, 0.0);
  double rearwingangle = GfParmGetNum(mCar->_carHandle, (char *)SECT_REARWING, (char *)PRM_WINGANGLE, (char*) NULL, 0.0);
  double frontwingarea = GfParmGetNum(mCar->_carHandle, (char *)SECT_FRNTWING, (char *)PRM_WINGAREA, (char*) NULL, 0.0);
  double frontwingangle = GfParmGetNum(mCar->_carHandle, (char *)SECT_FRNTWING, (char *)PRM_WINGANGLE, (char*) NULL, 0.0);
 double frontclift = GfParmGetNum(mCar->_carHandle, (char *)SECT_AERODYNAMICS, (char *)PRM_FCL, (char*) NULL, 0.0);
 double rearclift = GfParmGetNum(mCar->_carHandle, (char *)SECT_AERODYNAMICS, (char *)PRM_RCL, (char*) NULL, 0.0);
 double rearwingca = 1.23*rearwingarea*sin(rearwingangle);
 double frntwingca = 1.23*frontwingarea*sin(frontwingangle);

 double cl = frontclift + rearclift;
 double h = 0.0;

 int i;
 for (i = 0; i < 4; i++)
   {
     h += GfParmGetNum(mCar->_carHandle, WheelSect[i], (char *)PRM_RIDEHEIGHT, (char*) NULL, 0.20f);
   }
 h*= 1.5; h = h*h; h = h*h; h = 2.0 * exp(-3.0*h);
 CA = h*cl + 4.0*((frntwingca+rearwingca)/2);
 //fprintf(stderr,"CA=%.3f\n",CA);
 FCA = h*frontclift + 4.0*frntwingca;
 RCA = h*rearclift + 4.0*rearwingca;

}

/***************************/
// Compute aerodynamic drag coefficient CW.
void Driver::initCw()
{
  float cx = GfParmGetNum(mCar->_carHandle, (char *)SECT_AERODYNAMICS, (char *)PRM_CX, (char*) NULL, 0.0f);
  float frontarea = GfParmGetNum(mCar->_carHandle, (char *)SECT_AERODYNAMICS, (char *)PRM_FRNTAREA, (char*) NULL, 0.0f);
	CW = 0.645f*cx*frontarea;
}

/***************************/
void Driver::initCR()
{
 CR = GfParmGetNum(mCar->_carHandle, (char *)SECT_CAR, (char *)PRM_FRWEIGHTREP, (char *)NULL, 0.50);
}

/***************************/
// Init the friction coefficient of the the tires.
void Driver::initTireMu()
{
  char *WheelSect[4] = {(char *)SECT_FRNTRGTWHEEL, (char *)SECT_FRNTLFTWHEEL, (char *)SECT_REARRGTWHEEL, (char *)SECT_REARLFTWHEEL};
	float tm = FLT_MAX;
	int i;

	for (i = 0; i < 4; i++) {
	  tm = MIN(tm, GfParmGetNum(mCar->_carHandle, WheelSect[i], (char *)PRM_MU, (char*) NULL, 1.0f));
	}
	TIREMU = tm;
}

/***************************/
// Reduces the brake value such that it fits the speed (more downforce -> more braking).
float Driver::filterBrakeSpeed(float brake)
{
	float weight = (CARMASS + mCar->_fuel)*G;
	float maxForce = weight + CA*MAX_SPEED*MAX_SPEED;
	float force = weight + CA*currentspeedsqr;
	return brake*force/maxForce;
}

/***************************/
// Brake filter for pit stop.
float Driver::filterBPit(float brake)
{
	if (pit->getPitstop() && !pit->getInPit()) {
		float dl, dw;
		RtDistToPit(mCar, track, &dl, &dw);
		if (dl < PIT_BRAKE_AHEAD) {
			float mu = mCar->_trkPos.seg->surface->kFriction*TIREMU*PIT_MU;
			if (brakedist(0.0f, mu) > dl) {
				return 1.0f;
			}
		}
	}

	if (pit->getInPit()) {
		float s = pit->toSplineCoord(mCar->_distFromStartLine);
		// Pit entry.
		if (pit->getPitstop()) {
			float mu = mCar->_trkPos.seg->surface->kFriction*TIREMU*PIT_MU;

			if (s < pit->getNPitStart()) {
				// Brake to pit speed limit.
				float dist = pit->getNPitStart() - s;
				if (brakedist(pit->getSpeedlimit(), mu) > dist) {
					return 1.0f;
				}
			} else {
				// Hold speed limit.
				if (currentspeedsqr > pit->getSpeedlimitSqr()) {
					return pit->getSpeedLimitBrake(currentspeedsqr);
				}
			}
			// Brake into pit (speed limit 0.0 to stop)
			float dist = pit->getNPitLoc() - s;
			if (pit->isTimeout(dist)) {
				pit->setPitstop(false);
				return 0.0f;
			} else {
				if (brakedist(0.0f, mu) > dist) {
					return 1.0f;
				} else if (s > pit->getNPitLoc()) {
					// Stop in the pit.
			 		return 1.0f;
				}
			}
		} else {

			// Pit exit.
			if (s < pit->getNPitEnd()) {
				// Pit speed limit.
				if (currentspeedsqr > pit->getSpeedlimitSqr()) {
					return pit->getSpeedLimitBrake(currentspeedsqr);
				}
			}
		}
	}

	return brake;
}

/***************************/
// Brake filter for collision avoidance.
float Driver::filterBColl(float brake)
{
	if (simtime < 1.5)
		return brake;

	float mu = mCar->_trkPos.seg->surface->kFriction;
	int i;
	for (i = 0; i < opponents->getNOpponents(); i++) 
	{
		if (opponent[i].getState() & OPP_COLL) 
		{
			float ospeed = opponent[i].getSpeed();
			if (brakedist(ospeed, mu) + MIN(1.0, 0.5 + MAX(0.0, (getSpeed()-ospeed)/4)) > opponent[i].getDistance()) 
			{
				accelcmd = 0.0f;
				return 1.0f;
			}
		}
	}
	return brake;
}

/***************************/
// Antilocking filter for brakes.
float Driver::filterABS(float brake)
{
	if (mCar->_speed_x < ABS_MINSPEED) return brake;
	float origbrake = brake;
	float rearskid = MAX(0.0f, MAX(mCar->_skid[2], mCar->_skid[3]) - MAX(mCar->_skid[0], mCar->_skid[1]));
	int i;
	float slip = 0.0f;
	for (i = 0; i < 4; i++) {
		slip += mCar->_wheelSpinVel(i) * mCar->_wheelRadius(i);
	}
	slip *= 1.0f + MAX(rearskid, MAX(fabs(mCar->_yaw_rate)/5, fabs(angle)/6));
	slip = mCar->_speed_x - slip/4.0f;
	if (slip > abs_slip) {

		brake = brake - MIN(brake, (slip - abs_slip)/abs_range);
	}
	brake = MAX(brake, MIN(origbrake, 0.1f));
	return brake;
}

/***************************/
// TCL filter for accelerator pedal.
float Driver::filterTCL(float accel)
{
	accel = MIN(1.0f, accel);
	float accel1 = accel, accel2 = accel, accel3 = accel;

	if (mCar->_speed_x > 10.0f)
	{
		tTrackSeg *seg = mCar->_trkPos.seg;
		tTrackSeg *wseg0 = mCar->_wheelSeg(0);
		tTrackSeg *wseg1 = mCar->_wheelSeg(1);
	  int count = 0;

		if ((wseg0->surface->kRoughness > MAX(0.02, seg->surface->kRoughness*1.2) ||
		     wseg0->surface->kFriction < seg->surface->kFriction*0.8 ||
		     wseg0->surface->kRollRes > MAX(0.005, seg->surface->kRollRes*1.2)))
			count++;
		if ((wseg1->surface->kRoughness > MAX(0.02, seg->surface->kRoughness*1.2) ||
		     wseg1->surface->kFriction < seg->surface->kFriction*0.8 ||
		     wseg1->surface->kRollRes > MAX(0.005, seg->surface->kRollRes*1.2)))
			count++;

		if (count)
		{

			if (Mode != mode_normal &&
			    ((seg->type == TR_RGT && seg->radius <= 200.0f && mCar->_trkPos.toLeft < 3.0f) ||
			    (seg->type == TR_LFT && seg->radius <= 200.0f && mCar->_trkPos.toRight < 3.0f)))
				count++;
			accel1 = MAX(0.0f, MIN(accel1, (1.0f-(0.25f*count)) - MAX(0.0f, (getSpeed()-mCar->_speed_x)/10.0f)));

		}

		if (fabs(angle) > 1.0)
	 		accel1 = (float) MIN(accel1, 1.0f - (fabs(angle)-1.0)*1.3);
	}

	if (fabs(mCar->_steerCmd) > 0.02)
	{
		float decel = ((fabs(mCar->_steerCmd)-0.02f) * (1.0f+fabs(mCar->_steerCmd)) * 0.7f);
		accel2 = MIN(accel2, MAX(0.45f, 1.0f-decel));
	}

	float slip = (this->*GET_DRIVEN_WHEEL_SPEED)() - mCar->_speed_x;
	if (slip > tcl_slip) 
	  {

	    accel3 = accel3 - MIN(accel3, (slip - tcl_slip)/tcl_range);

	  }
	return MIN(accel1, MIN(accel2, accel3));
}

/***************************/
// Traction Control (TCL) setup.
void Driver::initTCLfilter()
{
  const char *traintype = GfParmGetStr(mCar->_carHandle, (char *)SECT_DRIVETRAIN, (char *)PRM_TYPE, (char *)VAL_TRANS_RWD);
	if (strcmp(traintype, VAL_TRANS_RWD) == 0) {
		GET_DRIVEN_WHEEL_SPEED = &Driver::filterTCL_RWD;
	} else if (strcmp(traintype, VAL_TRANS_FWD) == 0) {
		GET_DRIVEN_WHEEL_SPEED = &Driver::filterTCL_FWD;
	} else if (strcmp(traintype, VAL_TRANS_4WD) == 0) {
		GET_DRIVEN_WHEEL_SPEED = &Driver::filterTCL_4WD;
	}
}

/***************************/
// TCL filter plugin for rear wheel driven cars.
float Driver::filterTCL_RWD()
{
	return (mCar->_wheelSpinVel(REAR_RGT) + mCar->_wheelSpinVel(REAR_LFT)) *
			mCar->_wheelRadius(REAR_LFT) / 2.0f;
}

/***************************/
// TCL filter plugin for front wheel driven cars.
float Driver::filterTCL_FWD()
{
	return (mCar->_wheelSpinVel(FRNT_RGT) + mCar->_wheelSpinVel(FRNT_LFT)) *
			mCar->_wheelRadius(FRNT_LFT) / 2.0f;
}

/***************************/
// TCL filter plugin for all wheel driven cars.
float Driver::filterTCL_4WD()
{
	return ((mCar->_wheelSpinVel(FRNT_RGT) + mCar->_wheelSpinVel(FRNT_LFT)) *
			mCar->_wheelRadius(FRNT_LFT) +
		   (mCar->_wheelSpinVel(REAR_RGT) + mCar->_wheelSpinVel(REAR_LFT)) *
			mCar->_wheelRadius(REAR_LFT)) / 4.0f;
}

/***************************/
// Hold car on the track.
float Driver::filterTrk(float targetAngle)
{
  float angle = targetAngle;

  float tw = (mCar->_trkPos.seg->width - 2.0 * mCar->_dimension_y)/2.0f;
  //float tw2 = mCar->_trkPos.seg->width/2.0f;
  float tm = mCar->_trkPos.toMiddle;

  /* filter steering to stay on track */
  //if ( (fabs(tm) > tw) && (fabs(tm) < tw2) ) 
  if (fabs(tm) > tw)
    {
      float adjust;
      adjust = -2.5 * PI / 180.0 * tm / tw; 

      if ( fabs(adjust) > 5.0 * PI / 180.0 )
	{
	  if ( adjust > 0.0 )
	    adjust = 5.0 * PI / 180.0;
	  else
	    adjust = -5.0 * PI / 180.0;
	}
      NORM_PI_PI(angle);
      angle += adjust;
    }

  return angle;
}

/***************************/
// Compute the needed distance to brake.
float Driver::brakedist(float allowedspeed, float mu)
{
	float c = mu*G;
	float d = (CA*mu + CW)/car_Mass;
	float v1sqr = currentspeedsqr;
	float v2sqr = allowedspeed*allowedspeed;
	return -log((c + v2sqr*d)/(c + v1sqr*d))/(2.0f*d);
}


/*============== Print Parameters at Setup ===================*/
void Driver::showSetup()
{
  fprintf(stderr,"#====== Formulix SetUp=======\n");
  fprintf(stderr,"# %s in %s\n", mCar->_name, myCarName);

  if (showSegIndex) {
	fprintf(stderr,"# showSegIndex On\n");
  } else {
  	fprintf(stderr,"# showSegIndex Off\n");
  }
  if (testPitstop) {
	fprintf(stderr,"# testPitstop On\n");
  } else {
  	fprintf(stderr,"# testPitstop Off\n");
  }
  if (testQualifTime) {
	fprintf(stderr,"# testQualifTime On\n");
  } else {
  	fprintf(stderr,"# testQualifTime Off\n");
  }

  if (mode_verbose) {
	fprintf(stderr,"# modeVerbose On\n");
  } else {
  	fprintf(stderr,"# modeVerbose Off\n");
  }
  if (opp_verbose) {
	fprintf(stderr,"# oppVerbose On\n");
  } else {
  	fprintf(stderr,"# oppVerbose Off\n");
  }
  if (strat_verbose) {
	fprintf(stderr,"# stratVerbose On\n");
  } else {
  	fprintf(stderr,"# stratVerbose Off\n");
  }

  fprintf(stderr,"# \n");

  fprintf(stderr,"# fuelpermeter: %.5fl\n", fuel_PerMeter);
  fprintf(stderr,"# fuelperlap: %.2fl\n", fuel_PerLap);
  fprintf(stderr,"# fueltankfactor: %.1f\n", FUEL_FACTOR);
  fprintf(stderr,"# changeGear: %d\n", CHANGE_GEAR);
  if ( CHANGE_GEAR ) {
  	if (MAX_GEAR > 6) {
		fprintf(stderr,"# F1 Gear = %d\n", MAX_GEAR);
  	} else {
  		fprintf(stderr,"# car Gear = %d\n", MAX_GEAR);
  	}
  }
  fprintf(stderr,"# pitstrat: %d\n", pitStrat);
  fprintf(stderr,"# pitdamage: %d\n", maxDamage);
  
//  fprintf(stderr,"# \n");
//  fprintf(stderr,"# %s: FRONTCOLL_MARGIN=%.1f\n", car->_name, frontcollMargin);
//  fprintf(stderr,"# %s: SIDECOLL_MARGIN=%.1f\n", car->_name, sidecollMargin);

//  fprintf(stderr,"# %s: maxLeftMARGIN=%.1f\n", car->_name, maxleftMargin);
//  fprintf(stderr,"# %s: maxRightMARGIN=%.1f\n", car->_name, maxrightMargin);
  fprintf(stderr," \n");
}

/* -_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_--_-_-_*/


