/***************************************************************************

    file                 : driver.h
    created              : Thu Dec 20 01:20:19 CET 2002
    copyright            : (C) 2002-2004 Bernhard Wymann
    email                : berniw@bluewin.ch
    version              : $Id: driver.h,v 1.12 2006/03/06 22:43:50 berniw Exp $

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _DRIVER_H_
#define _DRIVER_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <tgf.h>
#include <track.h>
#include <car.h>
#include <raceman.h>
#include <robottools.h>
#include <robot.h>
//#include <portability.h>

#include "opponent.h"
#include "pit.h"
#include "strategy.h"
#include "cardata.h"
#include "raceline.h"
#include "line.h"


class Opponents;
class Opponent;
class Pit;
class AbstractStrategy;

enum { mode_normal=1, mode_avoiding, mode_correcting, mode_pitting, mode_stuck };
enum { avoidleft=1, avoidright=2, avoidside=4 };
enum { TEAM_MATE=1, TEAM_OPP };

class Line;

class Driver {
	public:
		Driver(int index);
		~Driver();

		/* Callback functions called from TORCS. */
		void initTrack(tTrack* t, void *carHandle, void **carParmHandle, tSituation *s);
		void newRace(tCarElt* car, tSituation *s);
		void drive();
		int pitCommand();
		void endRace(tSituation *s);

		tCarElt *getCarPtr() { return car; }
		tTrack *getTrackPtr() { return track; }
		float getSpeed() { return mycardata->getSpeedInTrackDirection(); /*speed;*/ }

	private:
		/* Utility functions */
  		void initCarSpec();
		void initCa();
		void initCw();
		void initCR();
		void initTireMu();
		void initTCLfilter();
		void clearCtrl();
		void updateTime();
		void update();
  		//void drvUpdate();
		void updateData();
		void setPitStatus();
		void stateStuck();
		void setControls();
		void updateStuck();
		void calcSpeed();
		void setMode( int newmode );
		void showSetup();

		float getAllowedSpeed(float lgfromstart);
		float getAccel();
		float getDistToSegEnd();
		float getBrake();
		int getGear();
		float getSteer(tSituation *s);
		float getClutch();
		vec2f getTargetPoint();
		//float getOffset();
		float brakedist(float allowedspeed, float mu);
		float smoothSteering( float steercmd );
		float correctSteering( float avoidsteer, float racesteer );
		float calcSteer( float targetAngle, int rl, float racesteer );
		void setAccelBrake(double speed, double barkefactor);
		void getMyOffset();
		void getOppState();
		void oppCollision();
		void oppLetPass();
		void getPitSpeed(double speed);
		bool onPitLimit();
		float getWidth() { return mycardata->getWidthOnTrack(); }
		
		float filterOverlap(float accel);
		float filterBColl(float brake);
		float filterABS(float brake);
		float filterBPit(float brake);
		float filterBrakeSpeed(float brake);
		float filterTurnSpeed(float brake);
		float filterTCL(float accel);
		float filterTrk(float accel);
		float filterSColl(float steer);
		float filterTCL_RWD();
		float filterTCL_FWD();
		float filterTCL_4WD();
		
		int isAlone();
	
		void computeRadius(int line, float *radius);
		
		//void loadSVG();
		//void saveSVG();

		// Setup param (loaded from xml file).
		bool saveSegData;
		bool saveRaceData;
  		bool showSegIndex;
  		bool drv_verbose;
  		bool mode_verbose;
		bool opp_verbose;
  		bool strat_verbose;
  		bool test_fuel;
  		bool testPitstop;
  		bool testQualifTime;
		bool drv_setup;
		int maxDamage;
		int pitStrat;
		float fuel_PerLap;
  		float fuel_PerMeter;
		double PitOffset;
		double pitEntryMargin;
		double pitExitMargin;
  		double pitEntrySpeed;
  		double pitExitSpeed;
		int CHANGE_GEAR;
		int MAX_GEAR;
		// Per robot global data.
		double currentsimtime;	// Store time to avoid useless updates.
		int Mode;
		bool avoidMode;
		int drvmode;
		int lastmode;
		int stuck;
  		int count_Stuck;
		bool isStuck;
		bool isWait;
		bool onColl;
		float speedangle;		// the angle of the speed vector relative to trackangle, > 0.0 points to right.
		float angle;
		double carMu;
		double carFriction;
		double ftank_Mass;		// Mass of full fuel tank.
		double car_Mass;		// Mass of car + fuel.
		double posFromStart;
		double carBorderdist;
		double carWalldist;
		bool car_goToWall;
		double carSpeed;
  		double maxCarSpeed;
  		double minCarSpeed;
		int maxCarDamage;
		int car_damage;
  		int car_damageLast;
  		int car_damageDiff;
		int car_damageLeft;
		double lastTimer;
  		bool stuckTimer;
		float myoffset;			// Offset to the track middle.
		float laststeer;
		float lastNSasteer;
		float stucksteer;
		
  		tSituation* s;
		tSituation* mSitu;
		tTrack* track;
		tTrack* mTrack;
		tCarElt *car;			// Pointer to tCarElt struct.
		tCarElt* mCar;       
		
		LRaceLine raceline;
		Line line;                     // Racing line finder

		Opponents *opponents;		// The container for opponents.
		Opponent *opponent;		// The array of opponents.

		Pit *pit;				// Pointer to the pit instance.
		AbstractStrategy *strategy;		// Pit stop strategy.

		static Cardata *cardata;		// Data about all cars shared by all instances.
		SingleCardata *mycardata;		// Pointer to "global" data about my car.

		static int curseg_id;
		static char *curseg_name;

		double simtime;       	// how long since the race started
		double avoidtime;    	// how long since we began avoiding
		double correcttimer; 	// how long we've been correcting
		double correctlimit; 	// level of divergence with raceline steering

		double brakedelay;
  		double CornerSpeed;
		double setAccel;
		double LetPass;
		int bumpCaution;

		double tcl_slip, tcl_range;
		double abs_slip, abs_range;
		float currentspeedsqr;		// Square of the current speed_x.
		float clutchtime;		// Clutch timer.
		float lastlookahead;		// Lookahead for steering in the previous step.
		float racesteer;     	// steer command to get to raceline
		float rlookahead;    	// how far ahead on the track we look for steering
		float raceoffset;    	// offset from middle of track towards which raceline is steering
		float avoidlftoffset; 	// closest opponent on the left
		float avoidrgtoffset; 	// closest opponent on the right
		float racespeed;     	// how fast raceline code says we should be going
		float avoidspeed;    	// how fast we should go if avoiding
		double rInverse;
		double leftOffset;
		double rightOffset;
		double incfactor;
		double rgtinc;
		double lftinc;
		float accelcmd, brakecmd;  
		v2d racetarget;      	// the 2d point the raceline is driving at.

		float *radius;
		int alone;
		int carindex;
		

		// Class constants.
		float G;
		float MAX_UNSTUCK_ANGLE;
		float UNSTUCK_TIME_LIMIT;
		float UNSTUCK2_TIME_LIMIT;
		float UNSTUCK3_TIME_LIMIT;
		float MAX_UNSTUCK_SPEED;
		float MIN_UNSTUCK_DIST;
		float FULL_ACCEL_MARGIN;
		float SHIFT;
		float SHIFT_UP;
		float SHIFT_DOWN;
		float SHIFT_MARGIN;
		float ABS_MINSPEED;
		float LOOKAHEAD_CONST;
		float LOOKAHEAD_FACTOR;
		float WIDTHDIV;
		double FRONTCOLL_MARGIN;
		double SIDECOLL_MARGIN;
		float BORDER_OVERTAKE_MARGIN;
		float OVERTAKE_OFFSET_SPEED;
		float PIT_LOOKAHEAD;
		float PIT_BRAKE_AHEAD;
		float PIT_MU;
		float MAX_SPEED;
		float CLUTCH_SPEED;
		float CENTERDIV;
		float DISTCUTOFF;
		float MAX_INC_FACTOR;
		float CATCH_FACTOR;
		float CLUTCH_FULL_MAX_TIME;
		float USE_LEARNED_OFFSET_RANGE;
		float TEAM_REAR_DIST;
		int MAX_DAMAGE_DIFF;
		// Data that should stay constant after first initialization.
		int INDEX;
		char* moduleName;	// Name of this module.
		char* trackName;	// Name of the track.
		char* myCarName;	// Name of the car.
		char carName[256];
		float MU_FACTOR;
		float MAXFUEL_PERMETER;
		double frontcoll_MARGIN;
  		double sidecoll_MARGIN;
		int MAX_UNSTUCK_COUNT;
		int MAX_UNSTUCK_COUNT2;
		int MAX_UNSTUCK_COUNT3;
		float ABS_SLIP;
		float ABS_RANGE ;
		float TCL_SLIP;
		float TCL_RANGE;
		double FUEL_FACTOR;	// fuel tank factor
		double WHEEL_BASE;  	// wheelbase of the car
		double CARMASS;		// Mass of the car only [kg].
		double BRAKE_PRESSURE;
		float CA;    		// Aerodynamic downforce coefficient.
		double FCA;  		// front downforce
		double RCA;  		// rear downforce
		double FWA;  		// front wing angle
		float CW;    		// Aerodynamic drag coefficient.
		double CR;   		// Front/rear weight repartition
		float TIREMU; 		// Friction coefficient of tires.
		float (Driver::*GET_DRIVEN_WHEEL_SPEED)();
		float OVERTAKE_OFFSET_INC;		// [m/timestep]
		
};

#endif // _DRIVER_H_

