/***************************************************************************

    file                 : opponent.cpp
    created              : Thu Apr 22 01:20:19 CET 2003
    copyright            : (C) 2003-2004 Bernhard Wymann
    email                : berniw@bluewin.ch
    version              : $Id: opponent.cpp,v 1.9 2006/03/06 22:43:50 berniw Exp $

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//#define OPP_DEBUG
#include "opponent.h"
#include "Vec2d.h"
#include "vardef.h"

// class variables and constants.
tTrack* Opponent::track;
const double Opponent::FRONTRANGEDIST = 200.0;			// [m] distance to check other cars in front me.
const double Opponent::BACKRANGEDIST = 100.0;			// [m] distance to check other cars behind me.
const float Opponent::LAP_BACK_TIME_PENALTY = -30.0f;		// [s]
const float Opponent::OVERLAP_WAIT_TIME = 5.0f;			// [s] overlaptimer must reach this time before we let the opponent pass.
const float Opponent::SPEED_PASS_MARGIN = 5.0f;			// [m/s] avoid overlapping opponents to stuck behind me.
const int Opponent::MIN_DAMAGE_DIFF = 1000;
const int Opponent::MAX_DAMAGE_DIFF = 3000;

Opponent::Opponent()
{
	teammate =false;
	teamMate = false;
	team = -1;
	MIN_DIST = 20.0;			// [m] min distance to compute it more accurate.
	FRONTCOLL_MARGIN = 1.50;		// [m] front safety margin.
	SIDECOLL_MARGIN = 1.0;			// [m] side safety margin.
}



void Opponent::update(tSituation *s, Driver *driver)
{
  //tCarElt *mycar = driver->getCarPtr();
  mycar = driver->getCarPtr();
  teamMate = !strncmp(car->_teamname, mycar->_teamname, 12);

  if (teamMate && team == -1) {
	fprintf(stderr,"%s TeamName = %s\n", mycar->_name, mycar->_teamname);
	if ( car->_name != mycar->_name) {
		fprintf(stderr,"%s TeamMate = %s\n", mycar->_name, car->_name);
	}
	team = TEAM_MATE;
  } else if (!teamMate) {
	team = TEAM_OPP;
  }

  // Set state
  initState();
  // If the car is out of the simulation ignore it.
  //if (car->_state & (RM_CAR_STATE_NO_SIMU & ~RM_CAR_STATE_PIT)) 
  if (car->_state & RM_CAR_STATE_NO_SIMU) 
  {
    return;
  }

  calcDist();

  speedangle = (float) -(cardata->getTrackangle() - atan2(car->_speed_Y, car->_speed_X));
  NORM_PI_PI(speedangle);

  mySpeed = driver->getSpeed();
  
  calcSpeed();
	oppSpeed = speed;
	

  calcState();

	if (!oppFaraway) // Opponent is in relevant range.
	{
		// Opponent aside.
		if (markASide) {
			sidedist = car->_trkPos.toMiddle - mycar->_trkPos.toMiddle;
			state |= OPP_SIDE;
		} 
		// Is opponent in front and slower.
		if (oppFront && !oppFaster) {
			state |= OPP_FRONT;

			if (teamFollow)
				state |= OPP_FRONT_FOLLOW;

			distance -= MAX(car->_dimension_x, mycar->_dimension_x);
			distance -= FRONTCOLL_MARGIN;

			// If the distance is small we compute it more accurate.
			if (distance < MIN_DIST) {
//
				straight2f carFrontLine(
					mycar->_corner_x(FRNT_LFT),
					mycar->_corner_y(FRNT_LFT),
					mycar->_corner_x(FRNT_RGT) - mycar->_corner_x(FRNT_LFT),
					mycar->_corner_y(FRNT_RGT) - mycar->_corner_y(FRNT_LFT)
				);

				float mindist = FLT_MAX;
				int i;
				for (i = 0; i < 4; i++) {
					vec2f corner(car->_corner_x(i), car->_corner_y(i));
					float dist = carFrontLine.dist(corner);
					if (dist < mindist) {
						mindist = dist;
					}
				}
//
				if (mindist < distance) {
					distance = mindist;
				}
			}

			catchdist = mySpeed*distance/(mySpeed - oppSpeed);

			float cardist = car->_trkPos.toMiddle - mycar->_trkPos.toMiddle;
			sidedist = cardist;
			cardist = fabs(cardist) - fabs(getWidth()/2.0f) - mycar->_dimension_y/2.0f;
                        //cardist = fabs(cardist) - mycar->_dimension_y;

			if (cardist < SIDECOLL_MARGIN) {
				state |= OPP_COLL;
			}
		}
		// Opponent is behind and faster.
	 	else if (oppBehind && oppSpeed > mySpeed - SPEED_PASS_MARGIN) {
			catchdist = mySpeed*distance/(oppSpeed - mySpeed);
			state |= OPP_BACK;
			distance -= MAX(car->_dimension_x, mycar->_dimension_x);
			distance -= FRONTCOLL_MARGIN;
		} 
		// Opponent is in front and faster.
		else if (oppFront && oppFaster) {
			state |= OPP_FRONT_FAST;
			distance -= MAX(car->_dimension_x, mycar->_dimension_x);
			if (teamFollow)
				state |= OPP_FRONT_FOLLOW;
			if (distance < MIN_DIST - (oppSpeed-mySpeed)*4)
				state |= OPP_FRONT;
		}
	}


  // Check if we should let overtake the opponent.
  updateOverlapTimer(s, mycar);

	if (!oppFaraway && oppLetPass && overlaptimer > OVERLAP_WAIT_TIME) {
		state |= OPP_LETPASS;
	}


  brakedistance = distance - car->_dimension_x;
}

void Opponent::initState()
{
  distance = DBL_MAX;
  state = OPP_IGNORE;
  oppBehind = false;
  oppFront = false;
  oppFaster = false;
  oppLetPass = false;
  oppFollow = false;
  oppFaraway = false;
  teamFollow = false;
  teamLetPass = false;
  markAsBack = false;
  markASide = false;
  markIsStuck = false;
}

// calculate distance
void Opponent::calcDist()
{
  //distance = (car->_trkPos.seg->lgfromstart + getDistToSegStart()) - mycar->_distFromStartLine;
  //float oppToStart = car->_trkPos.seg->lgfromstart + getDistToSegStart();
  //distance = oppToStart - mycar->_distFromStartLine;
  distance = car->_distFromStartLine - mycar->_distFromStartLine;
  	if (distance > track->length / 2.0) {
    		distance -= track->length;
  	} else if (distance < -track->length / 2.0) {
    		distance += track->length;
  	}
}

// Compute the length to the start of the segment.
float Opponent::getDistToSegStart()
{
	if (car->_trkPos.seg->type == TR_STR) {
		return car->_trkPos.toStart;
	} else {
		return car->_trkPos.toStart*car->_trkPos.seg->radius;
	}
}

void Opponent::calcSpeed()
{
  double angle;
  if (fabs(distance) < 20.0) {
  	angle = car->_yaw - mycar->_yaw;
    	NORM_PI_PI(angle);
    	if (fabs(angle) > 0.5) {
      		speed = getSpeedAngle(mycar->_yaw);
    	} else {
      		speed = car->_speed_x;
    	}
  } else {
    	speed = getSpeedAngle(RtTrackSideTgAngleL(&(car->_trkPos)));
  }
}

double Opponent::getSpeedAngle(double angle)
{
  Vec2d speed, dir;
  speed.x = car->_speed_X;
  speed.y = car->_speed_Y;
  dir.x = cos(angle);
  dir.y = sin(angle);
  return speed * dir;
}


void Opponent::calcState()
{
  //float MAXCARDIMX = MAX(car->_dimension_x, mycar->_dimension_x);
  float MAXCARDIMX = (car->_dimension_x / 2.0f) + (mycar->_dimension_x / 2.0f);
  //float COLLDIST = MAX(car->_dimension_x, mycar->_dimension_x);
  float COLLDIST = MAXCARDIMX * 0.98f;

  if (distance <= -BACKRANGEDIST || distance >= FRONTRANGEDIST) {
	oppFaraway = true;
	state = OPP_IGNORE;
   // Opponent in relevant range -BACK/+FRONT COLLDIST [m].
  } else if (distance > -BACKRANGEDIST && distance < FRONTRANGEDIST) {
	oppFaraway = false;
	if (team == TEAM_MATE) teamMate = true;
	/*
	// is opponent Stuck
    	if (oppSpeed < 0.2 && car->_gear <= 1) {
		markIsStuck = true;
	}
	*/
	// is opponent Back
    	if ((!teamMate && car->_remainingLaps > mycar->_remainingLaps) || 
    	  (teamMate && mycar->_dammage + MIN_DAMAGE_DIFF < car->_dammage && mycar->race.laps > 1)) {
      		markAsBack = true;
	}
	  // is opponent faster mycar
  	if (oppSpeed > mySpeed + 0.2) {
		oppFaster = true;
	} 
	  // is opponent aside
	if (distance > -COLLDIST && distance < COLLDIST) {
		markASide = true;	
	}
	  // Is opponent in front or behind
	if (distance > COLLDIST) {
		oppFront = true;
	} else if (distance < -COLLDIST) {
		oppBehind = true;
	}
	// Let TeamMate pass
	if ((teamMate && mycar->_dammage > car->_dammage + MIN_DAMAGE_DIFF && mycar->race.laps > 1) ||
    	  (teamMate && mycar->_dammage < car->_dammage + MIN_DAMAGE_DIFF && mycar->_pos > car->_pos && !markAsBack)) {
      		teamLetPass = true;
    	}
	// Let opponent pass
    	if ((!teamMate && car->_remainingLaps < mycar->_remainingLaps && oppFaster) || teamLetPass) {
		oppLetPass = true;
	}
	// Follow team mate
	if ( (teamMate && mycar->_dammage >= car->_dammage - MAX_DAMAGE_DIFF) && !markAsBack){
		teamFollow = true;
	}
  }
}


float Opponent::getCornerDist()
{
  straight2d frontLine(
    mycar->_corner_x(FRNT_LFT),
    mycar->_corner_y(FRNT_LFT),
    mycar->_corner_x(FRNT_RGT) - mycar->_corner_x(FRNT_LFT),
    mycar->_corner_y(FRNT_RGT) - mycar->_corner_y(FRNT_LFT)
  );
  straight2d rearLine(
    mycar->_corner_x(REAR_LFT),
    mycar->_corner_y(REAR_LFT),
    mycar->_corner_x(REAR_RGT) - mycar->_corner_x(REAR_LFT),
    mycar->_corner_y(REAR_RGT) - mycar->_corner_y(REAR_LFT)
  );
  straight2d leftLine(
    mycar->_corner_x(FRNT_LFT),
    mycar->_corner_y(FRNT_LFT),
    mycar->_corner_x(REAR_LFT) - mycar->_corner_x(FRNT_LFT),
    mycar->_corner_y(REAR_LFT) - mycar->_corner_y(FRNT_LFT)
  );
  straight2d rightLine(
    mycar->_corner_x(FRNT_RGT),
    mycar->_corner_y(FRNT_RGT),
    mycar->_corner_x(REAR_RGT) - mycar->_corner_x(FRNT_RGT),
    mycar->_corner_y(REAR_RGT) - mycar->_corner_y(FRNT_RGT)
  );
  double mindist = DBL_MAX;
  bool left[4];
  bool right[4];
  for (int i = 0; i < 4; i++) {
    Vec2d corner(car->_corner_x(i), car->_corner_y(i));
    double frontdist = frontLine.dist(corner);
    double reardist = rearLine.dist(corner);
    double leftdist = leftLine.dist(corner);
    double rightdist = rightLine.dist(corner);
    bool front = frontdist < reardist && reardist > mycar->_dimension_x ? true : false;
    bool rear = reardist < frontdist && frontdist > mycar->_dimension_x ? true : false;
    left[i] = leftdist < rightdist && rightdist > mycar->_dimension_y ? true : false;
    right[i] = rightdist < leftdist && leftdist > mycar->_dimension_y ? true : false;
    double dist = DBL_MAX;
    if (front) {
      dist = frontdist;
    } else if (rear) {
      dist = -reardist;
    }
    if (fabs(dist) < fabs(mindist)) {
      mindist = dist;
    }
  }
  if (fabs(mindist) > 3.0) {
    mindist -= SIGN(mindist) * 2.99;
  } else {
    mindist = SIGN(mindist) * 0.01;
  }
  bool lft = true;
  bool rgt = true;
  for (int j = 0; j < 4; j++) {
    if (!left[j]) {
      lft = false;
    }
  }
  for (int k = 0; k < 4; k++) {
    if (!right[k]) {
      rgt = false;
    }
  }
  if (lft || rgt) {
    // opponent aside
    mindist = 0.0;
  }
  return mindist;
}

#if 0
void Opponent::getOvtSideDist()
{
	double myCarTrkside = mycar->_trkPos.toMiddle;
	double oppCarTrkside = car->_trkPos.toMiddle;
	double m_carwidth =  mycar->_dimension_y/2.0 + fabs(getWidth()/2.0);
	double o_carwidth =  car->_dimension_y/2.0;

		double ovLeftDist = 0.0;
		double ovRightDist = 0.0;
		oppLetPass = false;
		sideDistToOvertake = 0.0;
		drv_line = drv_ovt;

			if (oppCarTrkside > 0) {
				ovLeftDist = car->_trkPos.seg->width/2 - (oppCarTrkside + o_carwidth);
				ovRightDist = -(car->_trkPos.seg->width/2);
			} else if (oppCarTrkside < 0) {
				ovLeftDist = car->_trkPos.seg->width/2;
				ovRightDist = -(car->_trkPos.seg->width/2 - (fabs(oppCarTrkside) + o_carwidth));
			}
			if (ovLeftDist > fabs(ovRightDist)) {
				sideDistToOvertake = ovLeftDist;
				drv_line = drv_left;
			} else {
				sideDistToOvertake = ovRightDist;
				drv_line = drv_right;
			}
			fprintf(stderr,"## %s can Overtake %s", mycar->_name, car->_name);
			if (oppCarTrkside > 0 && myCarTrkside > oppCarTrkside) {
				if ( ovLeftDist > m_carwidth ) {
					sideDistToOvertake = ovLeftDist;
					drv_line = drv_left;
					fprintf(stderr," to the left\n");
				} else {
					sideDistToOvertake = oppCarTrkside + fabs(ovRightDist) - o_carwidth;
					drv_line = drv_right;
					fprintf(stderr," to the right\n");
				}
			} else if (oppCarTrkside < 0 && myCarTrkside < oppCarTrkside) {
				if ( fabs(ovRightDist) > m_carwidth ) {
					sideDistToOvertake = ovRightDist;
					drv_line = drv_right;
					fprintf(stderr," to the right\n");
				} else {
					sideDistToOvertake = fabs(oppCarTrkside) + ovLeftDist - o_carwidth;
					drv_line = drv_left;
					fprintf(stderr," to the left\n");
				}
			} else {
				fprintf(stderr," ...[Error] check it?\n");
				if (!teamMate) {
					fprintf(stderr,"## %s ovLeftDist=%.2f ovRightDist=%.2f to %s (sideDistToOvertake=%.2f)\n", 
					mycar->_name, ovLeftDist, ovRightDist, car->_name, sideDistToOvertake);
				}
			}
}
#endif
double Opponent::getCarDist()
{
   double dist1 = car->_trkPos.toMiddle;
   double dist2 = mycar->_trkPos.toMiddle;
   double carcolldist = (mycar->_dimension_y / 2.0) + (car->_dimension_y / 2.0);
	if ((dist1 < 0 && dist2 > 0) || (dist1 > 0 && dist2 < 0))
		return (fabs(dist1) + fabs(dist2)) - carcolldist;
	if ((dist1 > 0 && dist2 > 0) || (dist1 < 0 && dist2 < 0))
		return fabs(dist1 - dist2) - carcolldist;
}

// Update overlaptimers of opponents.
void Opponent::updateOverlapTimer(tSituation *s, tCarElt *mycar)
{
	if (car->_remainingLaps <  mycar->_remainingLaps ||
			((teamMate) && mycar->_dammage > car->_dammage + MAX_DAMAGE_DIFF)) 
	{
		if (getState() & (OPP_BACK | OPP_SIDE)) {
			overlaptimer += s->deltaTime;
		} else if (getState() & OPP_FRONT) {
			overlaptimer = LAP_BACK_TIME_PENALTY;
		} else {
			if (overlaptimer > 0.0f) {
				if (getState() & OPP_FRONT_FAST) {
					overlaptimer = MIN(0.0f, overlaptimer);
				} else {
					overlaptimer -= s->deltaTime;
				}
			} else {
				overlaptimer += s->deltaTime;
			}
		}
	} else {
		overlaptimer = 0.0;
	}
}


// Initialize the list of opponents.
Opponents::Opponents(tSituation *s, Driver *driver, Cardata *c)
{
	opponent = new Opponent[s->_ncars - 1];
	int i, j = 0;
	for (i = 0; i < s->_ncars; i++) {
		if (s->cars[i] != driver->getCarPtr()) {
			opponent[j].setCarPtr(s->cars[i]);
			opponent[j].setCarDataPtr(c->findCar(s->cars[i]));
			opponent[j].setIndex(i);
			j++;
		}
	}
	Opponent::setTrackPtr(driver->getTrackPtr());
	nopponents = s->_ncars - 1;
}


Opponents::~Opponents()
{
	delete [] opponent;
}


void Opponents::update(tSituation *s, Driver *driver)
{
  int i;
	// Update opponent
	for (i = 0; i < s->_ncars - 1; i++) {
		opponent[i].update(s, driver);
	}
	// Set TeamMate
	for (i = 0; i < nopponents; i++) {
		 if (opponent[i].teamMate) {
			opponent[i].markAsTeamMate();
			break;	
		}
	}

}







