/***************************************************************************

    file                 : opponent.h
    created              : Thu Apr 22 01:20:19 CET 2003
    copyright            : (C) 2003-2004 Bernhard Wymann
    email                : berniw@bluewin.ch
    version              : $Id: opponent.h,v 1.8 2006/03/06 22:43:50 berniw Exp $

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _OPPONENT_H_
#define _OPPONENT_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <tgf.h>
#include <track.h>
#include <car.h>
#include <raceman.h>
#include <robottools.h>
#include <robot.h>
//#include <portability.h>

#include "driver.h"
#include "cardata.h"

class Driver;

// Opponent maintains the data for one opponent RELATIVE to the drivers car.
class Opponent {
	public:
		Opponent();

		void setCarPtr(tCarElt *car) { this->car = car; }
		void setCarDataPtr(SingleCardata *cardata) { this->cardata = cardata; }
		static void setTrackPtr(tTrack *track) { Opponent::track = track; }
		
		tCarElt *getCarPtr() { return car; }
		int getState() { return state; }
		float getCatchDist() { return catchdist; }
		float getDistance() { return distance; }
		float getBrakeDistance() { return brakedistance; }
		float getSideDist() { return sidedist; }
		float getWidth() { return cardata->getWidthOnTrack(); }
		float getSpeed() { return cardata->getSpeedInTrackDirection(); }
		//int getDrvLane() { return drv_line; }
		float getOverlapTimer() { return overlaptimer; }
		bool isTeamMate() { return teammate; }
		int getDamage() { return car->_dammage; }
		int getTeam() { return team; }
		void markAsTeamMate() { teammate = true; }
		void setIndex(int i) { index = i; }
		int getIndex() { return index; }
		float getSpeedAngle() { return speedangle; }
		
		void update(tSituation *s, Driver *driver);
		bool teamMate;

	private:
		float getDistToSegStart();
		void initState();
    		void calcState();
    		void calcDist();
    		void calcSpeed();
    		double getSpeedAngle(double angle);
    		float getCornerDist();
		void getOvtSideDist();
		double getCarDist();
		void updateOverlapTimer(tSituation *s, tCarElt *mycar);
		
		float distance;		// approximation of the real distance, negative if the opponent is behind.
		float brakedistance;	// distance minus opponent car length
		float catchdist;	// distance needed to catch the opponent (linear estimate).
		float sidedist;		// approx distance of center of gravity of the cars.
		int state;		// State variable to characterize the relation to the opponent, e. g. opponent is behind.
		bool teammate;		// Is this opponent a team mate of me?
		int team;		// counter to check Team Mate only the first time.
		int index;
		float overlaptimer;
		double speed;
		float oppSpeed;
		float mySpeed;
		float speedangle;
		double toMiddle;
    		double borderdist;
		double sideDistToOvertake;
	
		tCarElt *car;			// pointer to the opponents car   
    		tCarElt *mycar;  		// pointer to the my car
		SingleCardata *cardata;		// Pointer to global data about this opponent.
		
		bool markASide;
		bool markAsBack;
		bool markIsStuck;
    		bool oppFront;
		bool oppBehind;
		bool oppFaraway;
		bool oppFaster;
		bool oppFollow;
   		bool oppLetPass;
		bool teamLetPass;
		bool teamFollow;
		bool oppLetPassAccel;

		// class variables.
		static tTrack *track;

		// constants.
		double FRONTCOLL_MARGIN;
		double SIDECOLL_MARGIN;
		double MIN_DIST;
		static const double FRONTRANGEDIST;
		static const double BACKRANGEDIST;
		static const int MIN_DAMAGE_DIFF;
		static const int MAX_DAMAGE_DIFF;
		static const float SPEED_PASS_MARGIN;
		static const float LAP_BACK_TIME_PENALTY;
		static const float OVERLAP_WAIT_TIME;
};

// The Opponents class holds an array of all Opponents.
class Opponents {
	public:
		Opponents(tSituation *s, Driver *driver, Cardata *cardata);
		~Opponents();

		void update(tSituation *s, Driver *driver);
		Opponent *getOpponentPtr() { return opponent; }
		int getNOpponents() { return nopponents; }
		Opponent *opponent;
		int nopponents;

	private:
		Opponent* oppback;
    		Opponent* oppinrange;
   		Opponent* oppletpass;
};


#endif // _OPPONENT_H_
